#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


import unittest, tempfile, os, json

from src.mg_json_mgit_parser import exportToMgit, ProjectStructure
from src.mg_repo_info import MgRepoInfo

class TestMgitImportExport(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        os_handle, cls.mgitFname = tempfile.mkstemp(suffix='.mgit', prefix='test_mgit_', text=True)
        os.close(os_handle)
        # we only need the filename
        print('\nUsing temp filename: %s\n' % cls.mgitFname)


    def testEmptyExportImport(self):
        exportToMgit(self.mgitFname, '', [])
        proj = ProjectStructure()
        proj.fill_from_json_file(self.mgitFname, '.')
        self.assertEqual(proj.description, '')
        self.assertEqual(proj.repos, [])


    def testExportImportRepos(self):
        repo1 = MgRepoInfo('repo1', 'repo1', 'repo1')
        repo1.url = 'file://server/repo1.git'
        repo2 = MgRepoInfo('repo2', 'repo2', 'repo2')
        repo2.url = 'file://server/repo2.git'
        exportToMgit(self.mgitFname, 'some description', [repo2, repo1])

        with open(self.mgitFname) as f:
            d = json.load(f)

        # check that repos are sorted upon export
        self.assertEqual(d['repositories'][0]['destination'], 'repo1')


        proj = ProjectStructure()
        proj.fill_from_json_file(self.mgitFname, '.')

        self.assertEqual(proj.description, 'some description')
        self.assertEqual(len(proj.repos), 2)
        self.assertEqual(proj.repos[0].destination, 'repo1')
        self.assertEqual(proj.repos[1].destination, 'repo2')


    def testExportImportReposSnapshotMode(self):
        repo1 = MgRepoInfo('repo1', 'repo1', 'repo1')
        repo1.url = 'file://server/repo1.git'
        repo1.head = 'branch toto'
        repo1.commit_sha1 = '1111'

        repo2 = MgRepoInfo('Repo2', 'Repo2', 'Repo2')
        repo2.url = 'file://server/repo2.git'
        repo2.head = 'tag titi'
        repo2.commit_sha1 = '2222'

        repo3 = MgRepoInfo('repo3', 'repo3', 'repo3')
        repo3.url = 'file://server/repo3.git'
        repo3.head = 'commit 3333'
        repo3.commit_sha1 = '3333'

        exportToMgit(self.mgitFname, 'some description', [repo2, repo1, repo3], snapshotMode=False)

        proj = ProjectStructure()
        proj.fill_from_json_file(self.mgitFname, '.')

        self.assertEqual(proj.description, 'some description')
        self.assertEqual(len(proj.repos), 3)

        self.assertEqual(proj.repos[0].destination, 'repo1')
        self.assertEqual(proj.repos[0].head_type, 'branch')
        self.assertEqual(proj.repos[0].head, 'toto')

        self.assertEqual(proj.repos[1].destination, 'Repo2')
        self.assertEqual(proj.repos[1].head_type, 'tag')
        self.assertEqual(proj.repos[1].head, 'titi')

        self.assertEqual(proj.repos[2].destination, 'repo3')
        self.assertEqual(proj.repos[2].head_type, 'commit')
        self.assertEqual(proj.repos[2].head, '3333')


        exportToMgit(self.mgitFname, 'some description', [repo2, repo1, repo3], snapshotMode=True)

        proj = ProjectStructure()
        proj.fill_from_json_file(self.mgitFname, '.')

        self.assertEqual(proj.description, 'some description')
        self.assertEqual(len(proj.repos), 3)

        self.assertEqual(proj.repos[0].destination, 'repo1')
        self.assertEqual(proj.repos[0].head_type, 'commit')
        self.assertEqual(proj.repos[0].head, '1111')

        self.assertEqual(proj.repos[1].destination, 'Repo2')
        self.assertEqual(proj.repos[1].head_type, 'commit')
        self.assertEqual(proj.repos[1].head, '2222')

        self.assertEqual(proj.repos[2].destination, 'repo3')
        self.assertEqual(proj.repos[2].head_type, 'commit')
        self.assertEqual(proj.repos[2].head, '3333')





# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_git_switch_branch.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from src.mg_repo_tree import MgRepoTree


class Ui_GitSwitchBranch(object):
    def setupUi(self, GitSwitchBranch: QDialog) -> None:
        if not GitSwitchBranch.objectName():
            GitSwitchBranch.setObjectName(u"GitSwitchBranch")
        GitSwitchBranch.resize(883, 723)
        self.verticalLayout_2 = QVBoxLayout(GitSwitchBranch)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.labelBranchOrTag = QLabel(GitSwitchBranch)
        self.labelBranchOrTag.setObjectName(u"labelBranchOrTag")

        self.gridLayout.addWidget(self.labelBranchOrTag, 0, 1, 1, 1)

        self.lineEditBranchTagName = QLineEdit(GitSwitchBranch)
        self.lineEditBranchTagName.setObjectName(u"lineEditBranchTagName")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(2)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEditBranchTagName.sizePolicy().hasHeightForWidth())
        self.lineEditBranchTagName.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(10)
        self.lineEditBranchTagName.setFont(font)

        self.gridLayout.addWidget(self.lineEditBranchTagName, 0, 2, 1, 1)

        self.widget_2 = QWidget(GitSwitchBranch)
        self.widget_2.setObjectName(u"widget_2")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.widget_2.sizePolicy().hasHeightForWidth())
        self.widget_2.setSizePolicy(sizePolicy1)

        self.gridLayout.addWidget(self.widget_2, 0, 3, 1, 1)

        self.checkBoxDefaultForNotExist = QCheckBox(GitSwitchBranch)
        self.checkBoxDefaultForNotExist.setObjectName(u"checkBoxDefaultForNotExist")

        self.gridLayout.addWidget(self.checkBoxDefaultForNotExist, 1, 2, 1, 1)

        self.checkBoxDeleteLocalBranch = QCheckBox(GitSwitchBranch)
        self.checkBoxDeleteLocalBranch.setObjectName(u"checkBoxDeleteLocalBranch")
        self.checkBoxDeleteLocalBranch.setChecked(True)

        self.gridLayout.addWidget(self.checkBoxDeleteLocalBranch, 2, 2, 1, 1)

        self.checkBoxDeleteRemoteBranch = QCheckBox(GitSwitchBranch)
        self.checkBoxDeleteRemoteBranch.setObjectName(u"checkBoxDeleteRemoteBranch")
        self.checkBoxDeleteRemoteBranch.setChecked(True)

        self.gridLayout.addWidget(self.checkBoxDeleteRemoteBranch, 3, 2, 1, 1)

        self.widget_4 = QWidget(GitSwitchBranch)
        self.widget_4.setObjectName(u"widget_4")

        self.gridLayout.addWidget(self.widget_4, 4, 0, 1, 3)


        self.verticalLayout_2.addLayout(self.gridLayout)

        self.groupBoxBranchOrTagSelection = QGroupBox(GitSwitchBranch)
        self.groupBoxBranchOrTagSelection.setObjectName(u"groupBoxBranchOrTagSelection")
        self.groupBoxBranchOrTagSelection.setFont(font)
        self.verticalLayout_3 = QVBoxLayout(self.groupBoxBranchOrTagSelection)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_3 = QLabel(self.groupBoxBranchOrTagSelection)
        self.label_3.setObjectName(u"label_3")
        font1 = QFont()
        font1.setPointSize(9)
        self.label_3.setFont(font1)

        self.horizontalLayout_2.addWidget(self.label_3)

        self.label = QLabel(self.groupBoxBranchOrTagSelection)
        self.label.setObjectName(u"label")

        self.horizontalLayout_2.addWidget(self.label)

        self.lineEditBranchFilter = QLineEdit(self.groupBoxBranchOrTagSelection)
        self.lineEditBranchFilter.setObjectName(u"lineEditBranchFilter")
        sizePolicy2 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(9)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.lineEditBranchFilter.sizePolicy().hasHeightForWidth())
        self.lineEditBranchFilter.setSizePolicy(sizePolicy2)
        self.lineEditBranchFilter.setFont(font1)

        self.horizontalLayout_2.addWidget(self.lineEditBranchFilter)

        self.widget = QWidget(self.groupBoxBranchOrTagSelection)
        self.widget.setObjectName(u"widget")
        sizePolicy3 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy3.setHorizontalStretch(3)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.widget.sizePolicy().hasHeightForWidth())
        self.widget.setSizePolicy(sizePolicy3)

        self.horizontalLayout_2.addWidget(self.widget)

        self.pushButtonGrouping = QPushButton(self.groupBoxBranchOrTagSelection)
        self.pushButtonGrouping.setObjectName(u"pushButtonGrouping")
        sizePolicy4 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy4.setHorizontalStretch(0)
        sizePolicy4.setVerticalStretch(0)
        sizePolicy4.setHeightForWidth(self.pushButtonGrouping.sizePolicy().hasHeightForWidth())
        self.pushButtonGrouping.setSizePolicy(sizePolicy4)
        self.pushButtonGrouping.setFont(font1)

        self.horizontalLayout_2.addWidget(self.pushButtonGrouping)


        self.verticalLayout_3.addLayout(self.horizontalLayout_2)

        self.treeWidgetBranches = QTreeWidget(self.groupBoxBranchOrTagSelection)
        __qtreewidgetitem = QTreeWidgetItem()
        __qtreewidgetitem.setTextAlignment(0, Qt.AlignLeading|Qt.AlignVCenter);
        self.treeWidgetBranches.setHeaderItem(__qtreewidgetitem)
        QTreeWidgetItem(self.treeWidgetBranches)
        QTreeWidgetItem(self.treeWidgetBranches)
        QTreeWidgetItem(self.treeWidgetBranches)
        self.treeWidgetBranches.setObjectName(u"treeWidgetBranches")
        self.treeWidgetBranches.setFont(font1)
        self.treeWidgetBranches.setFrameShape(QFrame.NoFrame)
        self.treeWidgetBranches.setProperty("showDropIndicator", False)
        self.treeWidgetBranches.setAlternatingRowColors(False)
        self.treeWidgetBranches.setSortingEnabled(True)
        self.treeWidgetBranches.header().setVisible(True)

        self.verticalLayout_3.addWidget(self.treeWidgetBranches)


        self.verticalLayout_2.addWidget(self.groupBoxBranchOrTagSelection)

        self.widget_5 = QWidget(GitSwitchBranch)
        self.widget_5.setObjectName(u"widget_5")

        self.verticalLayout_2.addWidget(self.widget_5)

        self.widget_3 = QWidget(GitSwitchBranch)
        self.widget_3.setObjectName(u"widget_3")

        self.verticalLayout_2.addWidget(self.widget_3)

        self.groupBox_2 = QGroupBox(GitSwitchBranch)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.groupBox_2.setFont(font)
        self.verticalLayout = QVBoxLayout(self.groupBox_2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(23)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.labelRepoSelected = QLabel(self.groupBox_2)
        self.labelRepoSelected.setObjectName(u"labelRepoSelected")
        sizePolicy1.setHeightForWidth(self.labelRepoSelected.sizePolicy().hasHeightForWidth())
        self.labelRepoSelected.setSizePolicy(sizePolicy1)
        self.labelRepoSelected.setFont(font1)
        self.labelRepoSelected.setAlignment(Qt.AlignCenter)

        self.horizontalLayout.addWidget(self.labelRepoSelected)

        self.pushButtonAdjustRepoList = QPushButton(self.groupBox_2)
        self.pushButtonAdjustRepoList.setObjectName(u"pushButtonAdjustRepoList")
        sizePolicy5 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy5.setHorizontalStretch(1)
        sizePolicy5.setVerticalStretch(0)
        sizePolicy5.setHeightForWidth(self.pushButtonAdjustRepoList.sizePolicy().hasHeightForWidth())
        self.pushButtonAdjustRepoList.setSizePolicy(sizePolicy5)
        self.pushButtonAdjustRepoList.setFont(font1)

        self.horizontalLayout.addWidget(self.pushButtonAdjustRepoList)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.treeWidgetRepoList = MgRepoTree(self.groupBox_2)
        self.treeWidgetRepoList.setObjectName(u"treeWidgetRepoList")
        self.treeWidgetRepoList.setFont(font1)
        self.treeWidgetRepoList.setAlternatingRowColors(True)
        self.treeWidgetRepoList.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeWidgetRepoList.setRootIsDecorated(False)
        self.treeWidgetRepoList.setUniformRowHeights(True)
        self.treeWidgetRepoList.setItemsExpandable(False)
        self.treeWidgetRepoList.setSortingEnabled(True)

        self.verticalLayout.addWidget(self.treeWidgetRepoList)


        self.verticalLayout_2.addWidget(self.groupBox_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.buttonBox = QDialogButtonBox(GitSwitchBranch)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setFont(font1)
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)

        self.horizontalLayout_4.addWidget(self.buttonBox)


        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

        QWidget.setTabOrder(self.lineEditBranchTagName, self.checkBoxDefaultForNotExist)
        QWidget.setTabOrder(self.checkBoxDefaultForNotExist, self.checkBoxDeleteLocalBranch)
        QWidget.setTabOrder(self.checkBoxDeleteLocalBranch, self.checkBoxDeleteRemoteBranch)
        QWidget.setTabOrder(self.checkBoxDeleteRemoteBranch, self.lineEditBranchFilter)
        QWidget.setTabOrder(self.lineEditBranchFilter, self.pushButtonGrouping)
        QWidget.setTabOrder(self.pushButtonGrouping, self.treeWidgetBranches)
        QWidget.setTabOrder(self.treeWidgetBranches, self.pushButtonAdjustRepoList)
        QWidget.setTabOrder(self.pushButtonAdjustRepoList, self.treeWidgetRepoList)

        self.retranslateUi(GitSwitchBranch)
        self.buttonBox.accepted.connect(GitSwitchBranch.accept)
        self.buttonBox.rejected.connect(GitSwitchBranch.reject)

        QMetaObject.connectSlotsByName(GitSwitchBranch)
    # setupUi

    def retranslateUi(self, GitSwitchBranch: QDialog) -> None:
        GitSwitchBranch.setWindowTitle(QCoreApplication.translate("GitSwitchBranch", u"Git Switch branch", None))
        self.labelBranchOrTag.setText(QCoreApplication.translate("GitSwitchBranch", u"Choose branch", None))
        self.checkBoxDefaultForNotExist.setText(QCoreApplication.translate("GitSwitchBranch", u"If selected branch does not exist, switch to branch int", None))
        self.checkBoxDeleteLocalBranch.setText(QCoreApplication.translate("GitSwitchBranch", u"Delete local branch", None))
        self.checkBoxDeleteRemoteBranch.setText(QCoreApplication.translate("GitSwitchBranch", u"Delete remote branch", None))
        self.groupBoxBranchOrTagSelection.setTitle(QCoreApplication.translate("GitSwitchBranch", u"Branch selection", None))
        self.label_3.setText(QCoreApplication.translate("GitSwitchBranch", u"Display filter :", None))
        self.label.setText("")
        self.pushButtonGrouping.setText(QCoreApplication.translate("GitSwitchBranch", u"    Group          ", None))
        ___qtreewidgetitem = self.treeWidgetBranches.headerItem()
        ___qtreewidgetitem.setText(2, QCoreApplication.translate("GitSwitchBranch", u"Type of branch", None));
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("GitSwitchBranch", u"Present in repositories", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("GitSwitchBranch", u"Possible names", None));

        __sortingEnabled = self.treeWidgetBranches.isSortingEnabled()
        self.treeWidgetBranches.setSortingEnabled(False)
        ___qtreewidgetitem1 = self.treeWidgetBranches.topLevelItem(0)
        ___qtreewidgetitem1.setText(0, QCoreApplication.translate("GitSwitchBranch", u"sdfk", None));
        ___qtreewidgetitem2 = self.treeWidgetBranches.topLevelItem(1)
        ___qtreewidgetitem2.setText(1, QCoreApplication.translate("GitSwitchBranch", u"(in all 15)", None));
        ___qtreewidgetitem2.setText(0, QCoreApplication.translate("GitSwitchBranch", u"master", None));
        ___qtreewidgetitem3 = self.treeWidgetBranches.topLevelItem(2)
        ___qtreewidgetitem3.setText(1, QCoreApplication.translate("GitSwitchBranch", u"(in 13)", None));
        ___qtreewidgetitem3.setText(0, QCoreApplication.translate("GitSwitchBranch", u"feat/toto", None));
        self.treeWidgetBranches.setSortingEnabled(__sortingEnabled)

        self.groupBox_2.setTitle(QCoreApplication.translate("GitSwitchBranch", u"Targeted repositories", None))
        self.labelRepoSelected.setText(QCoreApplication.translate("GitSwitchBranch", u"3 repositories selected", None))
        self.pushButtonAdjustRepoList.setText(QCoreApplication.translate("GitSwitchBranch", u"Adjust Repository list", None))
        ___qtreewidgetitem4 = self.treeWidgetRepoList.headerItem()
        ___qtreewidgetitem4.setText(4, QCoreApplication.translate("GitSwitchBranch", u"Last Remote Syncho", None));
        ___qtreewidgetitem4.setText(3, QCoreApplication.translate("GitSwitchBranch", u"Status", None));
        ___qtreewidgetitem4.setText(2, QCoreApplication.translate("GitSwitchBranch", u"Head", None));
        ___qtreewidgetitem4.setText(1, QCoreApplication.translate("GitSwitchBranch", u"Git Repo Path", None));
        ___qtreewidgetitem4.setText(0, QCoreApplication.translate("GitSwitchBranch", u"hidden", None));
    # retranslateUi




# Generated from CHANGELOG.md on 2025-02-01 17:56:53.399206


content_html = '''
<h1 align="center">MultiGit</h1>

<p><em>by Phlipppe Fremy <a href="m&#x61;&#x69;&#108;&#x74;&#x6f;&#x3a;&#112;&#104;&#105;&#108;&#105;&#x70;&#x70;&#x65;&#46;&#102;&#x72;&#x65;&#x6d;&#x79;&#x40;&#x69;&#100;&#101;&#x6d;&#105;&#97;&#46;&#x63;&#111;&#x6d;">&#112;&#104;&#105;&#108;&#105;&#x70;&#x70;&#x65;&#46;&#102;&#x72;&#x65;&#x6d;&#x79;&#x40;&#x69;&#100;&#101;&#x6d;&#105;&#97;&#46;&#x63;&#111;&#x6d;</a></em></p>

<h2>Introduction</h2>

<p>Multigit help users to manage many Git repositories simultaneously.</p>

<h2>Version 1.6.1 (2025-02-01)</h2>

<p>Improvements</p>

<ul>
<li>Porting to Linux</li>
<li>Add support for <code><span style="background-color:#f5f5f5;">gitk</span></code> and <code><span style="background-color:#f5f5f5;">git gui</span></code></li>
<li>Add CI for Linux and Windows on Github</li>
<li>support cloning from url like: <code><span style="background-color:#f5f5f5;">fremy@host.xz:path/to/repo.git/</span></code></li>
</ul>

<h2>Version 1.6.0 (2023-04-14)</h2>

<p>First official release to the open source world!</p>

<p>Improvements:</p>

<ul>
<li>Using <code><span style="background-color:#f5f5f5;">Enter</span></code> on a repository triggers the same action as double-clicking it.</li>
<li>Better about dialog, listing all open source components and licenses</li>
<li>document multigit file format with a JSON schema</li>
</ul>

<p>Bug fixes:</p>

<ul>
<li>Fix problem in branch/tag checkout</li>
<li>it was impossible to create a branch with the name different only by case from another branch created in the past</li>
<li>Do not hide the list of repositories when filtering in dialog switch/delete branch</li>
<li>The settings dialog would not work when no config file exists</li>
<li>When cloning nested repositories, ensure that nested clone starts even if top clone fails</li>
</ul>

<h2>Version 1.5 (2022-12-14)</h2>

<p>New:</p>

<ul>
<li>Multigit uses multiple tabs</li>
<li>Clone multigit file and Apply multigit file now follow dependency order of nested git repositories</li>
<li>When checkouting a tag already on the current commit, force git to show the tag name</li>
<li>In git switch branch, git delete branch, git checkout tag, show in a tooltip which repositories contain a branch/tag</li>
<li>Show number of repositories and selected repositories in the status bar</li>
<li>Add option to auto-fetch on startup</li>
<li>Can choose repositories when using Git, custom command</li>
</ul>

<p>Bug fixes:</p>

<ul>
<li>Ensure JSON multigit file is sorted when exporting</li>
<li>No more "RunTimeError, C++ object already deleted"</li>
<li>Deals better with corrupted configuration file</li>
<li>Deals better with repositories deleted while showing the export to multigit dialog</li>
<li>No more exceptions when multiple instances access the same log file</li>
<li>avoids git complaining about branch with the same name as a directory by using argument '--' with checkout</li>
<li>Better handling of scrolling in git execution window</li>
<li>Better handling of abort requests</li>
</ul>

<h2>Version 1.4 (2022-06-03)</h2>

<p>New:</p>

<ul>
<li>Multigit now exists in command-line: mgitcmd</li>
<li>Branch without remote tracking branch: add tooltip for how to fix the situation</li>
<li>Git push tags: choose which tag to push</li>
<li>Git Revert: show the list of files being reverted</li>
<li>Git Delete Branch: better control on which branch to delete (local vs remote)</li>
<li>Git Delete Branch: can delete multiple branches</li>
<li>Git authentication failure: warn user when git encounters too many authentication
failures to avoid locking the user account.</li>
<li>Installation: automatically uninstall previous version before installing a new one</li>
</ul>

<p>Bug fixes:</p>

<ul>
<li>fix log not being written when configuration file is corrupted</li>
<li>handles correctly path with characters ] and [</li>
<li>better handling of git execution window</li>
<li>do not allow cloning of ssh repositories with empty user name</li>
<li>handle better the case where local branch can not be deleted because of unpushed commits</li>
<li>fix <em>Git Switch Branch</em> to deal with multiple origins</li>
<li>force Git to show the current tag when creating a new tag on an existing tag</li>
<li>warn user before exporting unusable mgit files (empty url, local branches)</li>
</ul>

<h2>Version 1.3 (2022-01-27)</h2>

<p>New funtionality</p>

<ul>
<li>Apply a multigit file to a directory</li>
<li>When exporting to a multigit file, choice between exporting branches or a snapshot</li>
<li>New column with remote URL (can be hidden)</li>
<li>Configure visibility of SHA1 and URL column, of bottom tab Last Commit and Modified Files</li>
<li>Git swtich can switch to branch and (new) to tag</li>
</ul>

<p>Various bug fixes and small improvements:  </p>

<ul>
<li>fix username handling on clone url</li>
<li>fix CSV and Multigit export for unicode branch names and repositories names</li>
<li>fix sorting of repositories in clone/apply/export, make it case-insensitive</li>
<li>switch branch, when branches are grouped by name, correctly display the selected branch name</li>
<li>deal better with emtpy URL in multigit file</li>
<li>branch filter is now case-insensitive</li>
</ul>

<h2>Version 1.2 (2021-12-20)</h2>

<p>Main changes are:</p>

<ul>
<li>Menu to copy repository informations: url, branch name, tag, sha1</li>
<li>Add buttons in main interface to switch base directory quickly</li>
<li><p>When cloning over existing directories, tool can delete the directories or skip them</p></li>
<li><p>Details of the changes</p>

<ul>
<li><p>Bug correction</p>

<ul>
<li>support branch names using unicode characters</li>
<li>Export to JSON and CSV is not done in alphabetical order </li>
<li>Tool shows error message when a repository is deleted</li>
</ul></li>
<li><p>Improvements</p>

<ul>
<li>Better handling of existing directories when cloning</li>
<li>Add menu to copy SHA1, remote URL or full path</li>
<li>Open or switch base directory directly from main window</li>
<li>Better handling of existing directories when cloning</li>
<li>add ability to limit the maximum number of parallel git processes running</li>
<li>log git commands to a more readable file</li>
<li>log of git execution window can be copied to clipboard</li>
<li>Show the same menu in the menu bar and in the right-mouse click menu</li>
<li>Ask user to confirm when launching too many TortoiseGit actions</li>
<li>Git commit can recall previous commit messages</li>
<li>New column with SHA1</li>
<li>Add tooltip description to all menu items</li>
</ul></li>
</ul></li>
</ul>

<h2>Version 1.1.0 (2021-12-09)</h2>

<ul>
<li><p>Bug correction</p>

<ul>
<li>Git actions failed on large amount of repositories</li>
<li>Error on tracking remote repository for new created branch</li>
</ul></li>
<li><p>Improvements</p>

<ul>
<li>progress bar when exporting to mulitgit file</li>
<li>Add tortoise git command to switch/checkout branch</li>
<li>Add Git switch branch command</li>
<li>Add Git delete branch command</li>
<li>Add Git commit command</li>
</ul></li>
</ul>

<h2>Version 1.0.0 (2021-04-19)</h2>

<ul>
<li>Project Configuration File
<ul>
<li>username removed from url</li>
<li>username used to clone is specified in UI</li>
</ul></li>
<li>Git programs need to be individually activated to be available</li>
<li>Fix symlink management when opening directory and exporting project configuration file</li>
<li>Fix git switch: replace git switch (not backward compatible) by git checkout</li>
</ul>

<h2>Version 0.8.0 (2021-04-09)</h2>

<ul>
<li>Clone from configuration file
<ul>
<li>Improve cloning mecanism</li>
<li>fix if it is the first time (no last file)</li>
</ul></li>
</ul>

<h2>Version 0.7.1 (2021-04-07)</h2>

<ul>
<li>Fix TortoiseGit problem for repositories with space</li>
<li>Speedup csv export by filling data asynchronously</li>
<li>Fix export of empty repository</li>
<li>Git execution window:
<ul>
<li>Improving handling of abort</li>
<li>Display progressive clone output  </li>
</ul></li>
</ul>

<h2>Version 0.7.0 (2021-03-31)</h2>

<ul>
<li>Bugfix: support for symlinks in working dir</li>
<li>Create Project folder from multigit json configuration file (File menu)
<ul>
<li>List repositories </li>
<li>Analyze the json file</li>
<li>Build the project in the specified directory</li>
</ul></li>
<li>Generate multigit json configuration file from opened directory (File menu)</li>
</ul>

<h2>Version 0.6.8 (2021-02-26)</h2>

<ul>
<li>use a dedicated column and icon for showing that refresh is in progress</li>
<li>refresh the selected repositories when switching back to window</li>
<li>can run a custom git command</li>
<li>git command execution: use fixed font and better indentation of results</li>
<li>export CSV (warning, minor change of format)
<ul>
<li>can choose which fields to export</li>
<li>one column added to export the commit date</li>
</ul></li>
<li>bugfix: ESC during CSV export would crash</li>
<li>columns "status" and "last remote synchro" are sorted using natural sort</li>
</ul>

<h2>Version 0.6.7 (2021-01-27)</h2>

<ul>
<li>Show README and last changelog directly in the tool</li>
<li>Refresh repository after closing Sublime Text or Source Tree</li>
<li>Trigger the default action on a repo with enter</li>
<li>limit size of log debug file</li>
<li>shortcut F6 to fetch all</li>
<li>bugfix: copy/paste icon from properties does not show up !</li>
<li>fix bug: refresh all repo on F5 and not just the repo added</li>
<li>fix bug: wrong repo opened by sublimemerge and sourcetree</li>
<li>fix bug: exception when launching sublime merge and sourcetree</li>
<li>refresh repository view automatically after execution of TortoiseGit or Git Bash</li>
<li>fix bug: after launching TortoiseGit, an exception was sometimes raised, showing an exception dialog</li>
</ul>

<h2>Version 0.6.2 (2020-09-15)</h2>

<ul>
<li>add git push --tags.</li>
<li>mypy compliant 100%</li>
<li>better handling of repos added and removed</li>
<li>parallel execution of git jobs</li>
</ul>

<h2>Version 0.6 (2020-07-22)</h2>

<ul>
<li>ability to create annotated tags.</li>
<li>add Git Bash to the Git menu.</li>
<li>add Tortoise Git Diff to menu.</li>
<li>validation dialog before issuing git revert.</li>
<li>export all repository status to a CSV file</li>
<li>use colors for branch and tags.</li>
<li>synchronise the view asynchronously</li>
<li>remember last used repository upon startup</li>
<li>small user interface improvements</li>
<li>fixed a few bugs</li>
</ul>

<h2>Version 0.5 (2021-03-01)</h2>

<ul>
<li>Fix execution from a directory with no access rights</li>
<li>Configure action for double-click.</li>
<li>Autocheck availability of TortoiseGit</li>
<li>Disable TortoiseGit, SourceTree and SublimeMerge if not available</li>
<li>Other small bug fixes</li>
</ul>

<h2>Version 0.4 (2020-01-07)</h2>

<ul>
<li>Can now launch some git commands directly on the repo: fetch, pull, push, revert, tag</li>
<li>Add icons for Git and Tortoise Git operations</li>
<li>Add copy button to properties</li>
<li>add About dialog</li>
<li>Add "show in explorer" functionality</li>
<li>In repo properties, show all tags pointing to a given commit</li>
<li>Show the synchro status with remote branche</li>
<li>Add preference dialog to configure location of Git and TortoiseGit</li>
<li>Deal better with the absence of Git or TortoiseGit</li>
</ul>

<h2>Version 0.3 (2019-11-27)</h2>

<ul>
<li>Add TortoiseGit actions on repositories</li>
<li>Add a context menu</li>
<li>Fix refresh action</li>
<li>Add an icon</li>
<li>limit size of the diff</li>
<li>deal better with git diff generating invalid encodings</li>
<li>colorize diff</li>
</ul>

<h2>Version 0.2 (2019-09-12)</h2>

<ul>
<li>show last commit of the selected repo</li>
<li>show diff status of modified files of the selected repo</li>
<li>add property dialog</li>
<li>avoid detection of non-git directories</li>
</ul>

<h2>Version 0.1 (2019-08-23)</h2>

<ul>
<li>list all repos in a given directory, with their branch name and tags</li>
</ul>

'''



#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#

from typing import Optional

import platform

from PySide2.QtWidgets import QDialog, QWidget
from PySide2.QtCore import Qt
from PySide2.QtCore import qVersion


from src.gui.ui_about import Ui_dialogAbout
from src.gui.ui_about_license import Ui_FullLicenseInfoDialog
from src.gui.content_full_license_info import content_html
from src.mg_const import VERSION

MULTIGIT_VERSION_MARKER = '[version]'

class MgAboutDialog(QDialog):

    def __init__(self, parent: Optional[QWidget]) -> None:
        super().__init__(parent)
        self.setModal(False)
        # noinspection PyTypeChecker
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self.ui = Ui_dialogAbout()
        self.ui.setupUi(self)

        self.ui.pushButtonOk.clicked.connect(self.close)
        self.ui.pushButtonFullLicense.clicked.connect(self.showFullLicenseInfo)
        htmlContent = self.ui.textBrowserContent.toHtml()
        if MULTIGIT_VERSION_MARKER in htmlContent:
            htmlContent = htmlContent.replace(MULTIGIT_VERSION_MARKER, VERSION)

        htmlContent = htmlContent.replace('Python 3.8', f'Python {platform.python_version()}')
        htmlContent = htmlContent.replace('Qt for Python 5.15', f'Qt for Python {qVersion()}')

        self.ui.textBrowserContent.setHtml(htmlContent.replace(MULTIGIT_VERSION_MARKER, VERSION))





    def showFullLicenseInfo(self) -> None:
        dlg = QDialog(self)
        dlg.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        ui = Ui_FullLicenseInfoDialog()
        ui.setupUi(dlg)
        ui.textBrowser.setHtml(content_html)
        dlg.show()


def showDialogAbout(parent: Optional[QWidget]) -> None:
    '''Show the about dialog'''
    dlg = MgAboutDialog(parent)
    dlg.exec()




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import Optional

from PySide2.QtWidgets import QWidget, QDialog, QTextEdit, QPushButton, QVBoxLayout, QHBoxLayout, QFrame
from PySide2.QtCore import Qt

from src.gui.content_whatisnew import content_html
from src.mg_const import *
import src.mg_config as mgc


def showWhatisnewIfAppropriate() -> None:
    '''Show the 'What is new' dialog if it was never shown for the current version of MultiGit'''
    last_shown = mgc.get_config_instance()[mgc.CONFIG_LAST_SHOWN_WHATISNEW]
    if last_shown is not None and last_shown < VERSION:
        showWhatIsNew()
    mgc.get_config_instance()[mgc.CONFIG_LAST_SHOWN_WHATISNEW] = VERSION


def showWhatIsNew(parent: Optional[QWidget] = None) -> None:
    '''Show what's new dialog'''
    dialog = QDialog(parent)
    dialog.setWindowTitle("What's New")
    dialog.setWindowFlags( dialog.windowFlags() & ~Qt.WindowContextHelpButtonHint)

    text = QTextEdit(dialog)
    text.setReadOnly(True)
    text.setHtml(content_html)
    text.setFrameShape( QFrame.NoFrame )
    text.zoomIn(2)

    okButton = QPushButton('OK', dialog)
    okButton.clicked.connect( dialog.accept )

    vlayout = QVBoxLayout(dialog)
    vlayout.addWidget(text)
    hbutlayout = QHBoxLayout(dialog)
    hbutlayout.addStretch()
    hbutlayout.addWidget(okButton)
    vlayout.addLayout(hbutlayout)

    dialog.resize(700, 700)
    dialog.exec()



#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import cast, List, Optional, Callable, Any, Union, Literal
import subprocess, functools, logging, pathlib

from PySide2.QtWidgets import QMainWindow, QFileDialog, QMessageBox, QAction, QApplication, QLineEdit, QTabBar
from PySide2.QtCore import QTimer, Qt, QSignalMapper
from PySide2.QtGui import QCloseEvent

from src.gui.ui_main_window import Ui_MainWindow
from src.mg_actions import MgActions
from src.mg_dialog_whatisnew import showWhatIsNew, showWhatisnewIfAppropriate
from src.mg_dialog_export_mgit import runDialogExportMgit
from src.mg_dialog_export_csv import runDialogExportCsv
from src.mg_dialog_clone_from_mgit import runDialogCloneFromMgitFile
from src.mg_dialog_apply_mgit_file import runDialogApplyMgitFile
from src.mg_dialog_about import showDialogAbout
from src.mg_plugin_mgr import pluginMgrInstance
from src.mg_multigit_widget import MgMultigitWidget
from src.mg_repo_tree_item import MgRepoTreeItem
from src.mg_repo_tree import MgRepoTree
from src.mg_tools import ExecExplorer, ExecGit
from src.mg_dialog_settings import runDialogEditSettings
from src.mg_exec_window import MgExecWindow
from src import mg_config as mgc
from src.mg_const import VERSION, DISPLAY_FETCH_ON_STARTUP_COUNTDOWN_INIT


logger = logging.getLogger('mg_main_window')
dbg = logger.debug
info = logger.info


class MgMainWindow(QMainWindow, Ui_MainWindow):

    def __init__(self, argv: List[str] ) -> None:
        super().__init__()
        self.refreshOnFocusBack = True
        self.setupUi(self)

        self.config = mgc.get_config_instance()
        self.gitIsOk = False
        self.argv = argv

        if self.config[mgc.CONFIG_MAINWINDOW_GEOMETRY]:
            (x, y, w, h) = self.config[mgc.CONFIG_MAINWINDOW_GEOMETRY]
            self.setGeometry(x, y, w, h)
        else:
            self.setGeometry(self.x(), self.y(), 1000, 700)  # default reasonable window size when no configuration

        self.setupActions()
        self.setupMenus()
        self.setupConnections()
        pluginMgrInstance.setupTopMenu(self)

        self.setWindowTitle('MultiGit OpenSource v{}'.format(VERSION))

        QTimer.singleShot(0, self.checkGitOkAndOpenDefaultRepo)
        QTimer.singleShot(0, showWhatisnewIfAppropriate)


    def currentMultigit(self) -> MgMultigitWidget:
        '''Return the MgMultigitWidget active in the current tab'''
        # Note: in theory, self.tabRepos.currentWidget() can be None, when no tab have been created.
        #       However, in practice in our code, creating one or multiple tabs is the first thing we do
        #       after showing the window. So, we consider that it is never None
        return cast(MgMultigitWidget, self.tabRepos.currentWidget())


    def multigitWidgetFromIdx(self, idx: int) -> MgMultigitWidget:
        '''Return the MgMultigitWidget active in the current tab'''
        return cast(MgMultigitWidget, self.tabRepos.widget(idx))


##########################################################################
#
#       Methods related to the display updates and internal machinery
#
##########################################################################

    def setupActions(self) -> None:
        self.mgActions = MgActions(self)

        self.mgActions.actionGitFetchAll = QAction(self)
        self.mgActions.actionGitFetchAll.setText(u"Fetch All (current tab)")
        self.mgActions.actionGitFetchAll.setToolTip("Run <code>git fetch</code> on all repositories of the current tab")
        self.mgActions.actionGitFetchAll.setShortcut("F6")

        self.mgActions.actionGitFetchAllOnAllTabs = QAction(self)
        self.mgActions.actionGitFetchAllOnAllTabs.setText(u"Fetch All (all tabs)")
        self.mgActions.actionGitFetchAllOnAllTabs.setToolTip("Run <code>git fetch</code> on all repositories of all tabs")
        self.mgActions.actionGitFetchAllOnAllTabs.setShortcut("F7")


    def setupMenus(self) -> None:
        '''Setup all menu which were not already prepared by the Designer UI'''
        # menu File
        self.fillRecentDirMenu()
        self.mgActions.setupMenuView(self.menuView)

        # menu git
        self.menuGit.addAction(self.mgActions.actionGitFetchAll)
        self.menuGit.addAction(self.mgActions.actionGitFetchAllOnAllTabs)
        self.menuGit.addSeparator()
        self.mgActions.setupMenuGit(self.menuGit)

        # menu git programs
        self.mgActions.setupMenuGitPrograms(self.menuGitPrograms, includeConfigureGitPrograms=True)


    def fillRecentDirMenu(self) -> None:
        dbg('fillRecentDirMenu()')
        self.menuOpenRecentDirectory.clear()
        for dirname in self.config.lruAsList(mgc.CONFIG_LAST_OPENED):    # type: str
            action = QAction(dirname, self.menuOpenRecentDirectory)
            action.setData( dirname )
            self.menuOpenRecentDirectory.addAction( action )


    def dispatchToActiveMultigitTab(self, methodName: str) -> Callable[[], None]:
        '''Return a function which calls the methodName on the active tab widget
        (at the time where the method is called)'''
        def callActiveMultigitTabMethod() -> None:
            targetMethod = getattr(self.currentMultigit(), methodName)
            targetMethod()

        return callActiveMultigitTabMethod


    def dispatchToActiveTreeOfMultigitTab(self, methodName: str, with_arg1: bool = False) -> Union[Callable[[], None], Callable[[Any], None]]:
        '''Return a function which calls the methodName on all tab widget
        (at the time where the method is called).

        The method may support an argument if with_arg1 is True
        '''
        callActiveTreeOfMultigitTabMethod: Union[Callable[[], None], Callable[[Any], None]]
        if with_arg1:
            def callActiveTreeOfMultigitTabMethod(arg1: Any) -> None:
                targetMethod = getattr(self.currentMultigit().repoTree, methodName)
                targetMethod(arg1)

        else:
            def callActiveTreeOfMultigitTabMethod() -> None:
                targetMethod = getattr(self.currentMultigit().repoTree, methodName)
                targetMethod()

        return callActiveTreeOfMultigitTabMethod


    def allTreeOfMultigitTab(self) -> List[MgRepoTree]:
        return [
            cast(MgMultigitWidget, self.tabRepos.widget(tabIdx)).repoTree
            for tabIdx in range(self.tabRepos.count())
        ]


    def setupConnections(self) -> None:
        '''Setup all actions and corresponding slots'''

        ### menu File
        self.actionOpenDirectory.triggered.connect(self.slotOpenDir)
        self.menuOpenRecentDirectory.triggered.connect(self.slotOpenRecentDirTriggered)
        self.actionExportCSV.triggered.connect(self.slotExportCsv)
        self.actionEditPreferences.triggered.connect( self.slotEditSettings )
        # ----
        self.actionAddTab.triggered.connect( lambda checked: self.slotAddTab() )  # note: ignore the checked parameter
        self.actionDupTab.triggered.connect( self.slotDupTab )
        self.actionRenameTab.triggered.connect( self.slotRenameTab )
        self.actionCloseTab.triggered.connect( self.slotCloseTab )
        # ----
        self.actionOpenProject.triggered.connect(self.slotCloneFromMgitFile)
        self.actionApplyMultigitFile.triggered.connect(self.slotApplyMgitFile)
        self.actionExportToMgit.triggered.connect(self.slotExportToMgit)
        self.actionQuit.triggered.connect( self.close )

        ### menu View
        self.actionViewLastCommit.setChecked(mgc.get_config_instance().get(mgc.CONFIG_VIEW_TAB_LAST_COMMIT, False))
        self.actionViewLastCommit.changed.connect(self.slotViewTabChanged)
        self.actionViewModifiedFiles.setChecked(mgc.get_config_instance().get(mgc.CONFIG_VIEW_TAB_MOD_FILES, False))
        self.actionViewModifiedFiles.changed.connect(self.slotViewTabChanged)

        # trigger the slot once to setup the GUI
        self.actionViewColSha1.setChecked(mgc.get_config_instance().get(mgc.CONFIG_VIEW_COL_SHA1, True))
        self.actionViewColSha1.changed.connect(self.slotViewColSha1Changed)
        self.actionViewColURL.setChecked(mgc.get_config_instance().get(mgc.CONFIG_VIEW_COL_URL, True))
        self.actionViewColURL.changed.connect(self.slotViewColUrlChanged)
        self.actionRefreshAll.triggered.connect(self.dispatchToActiveMultigitTab('slotRefreshAll'))

        self.mgActions.actionRefreshSelected.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotRefreshSelected'))
        self.mgActions.actionShowInExplorer.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotShowInExplorer'))
        self.mgActions.menuCopy.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotMenuCopyAction', True))
        self.mgActions.menuCopy.aboutToShow.connect(self.slotMenuCopyAboutToShow)
        self.mgActions.menuCopy.aboutToHide.connect(self.slotMenuCopyAboutToHide)


        ### menu Git
        self.mgActions.actionGitFetchAll.triggered.connect(self.dispatchToActiveMultigitTab('slotGitFetchAll'))
        self.mgActions.actionGitFetchAllOnAllTabs.triggered.connect(self.slotFetchAllReposOnAllTabs)
        self.mgActions.actionGitProperties.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitProperties'))
        self.mgActions.actionGitCommit.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitCommit'))
        self.mgActions.actionGitRevert.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitRevert'))
        self.mgActions.actionGitPush.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitPush'))
        self.mgActions.actionGitPushTag.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitPushTag'))
        self.mgActions.actionGitPull.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitPull'))
        self.mgActions.actionGitFetch.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitFetch'))
        self.mgActions.actionGitTag.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitTag'))
        self.mgActions.actionGitCreateBranch.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitCreateBranch'))
        self.mgActions.actionGitSwitchBranch.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitSwitchBranch'))
        self.mgActions.actionGitCheckoutTag.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitCheckoutTag'))
        self.mgActions.actionGitDeleteBranch.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitDeleteBranch'))
        self.mgActions.actionGitBash.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitBash'))
        self.mgActions.actionGitRunCommand.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitRunCommand'))

        # Menu Git programs
        self.mgActions.actionSourceTree.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotSourcetree'))
        self.mgActions.actionSublimeMerge.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotSublimemerge'))
        self.mgActions.actionGitGui.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitGui'))
        self.mgActions.actionGitK.triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotGitK'))

        self.mgActions.actionTGitShowLog  .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitShowLog'))
        self.mgActions.actionTGitCommit   .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitCommit'))
        self.mgActions.actionTGitDiff     .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitDiff'))
        self.mgActions.actionTGitRevert   .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitRevert'))
        self.mgActions.actionTGitPush     .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitPush'))
        self.mgActions.actionTGitPull     .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitPull'))
        self.mgActions.actionTGitFetch    .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitFetch'))
        self.mgActions.actionTGitTag      .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitTag'))
        self.mgActions.actionTGitSwitch   .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitSwitch'))
        self.mgActions.actionTGitBranch   .triggered.connect(self.dispatchToActiveTreeOfMultigitTab('slotTGitBranch'))

        self.mgActions.actionConfigureGitProgram.triggered.connect(self.slotEditSettingsGitProgram)

        # menu About
        self.actionWhatIsNew.triggered.connect( self.slotWhatIsNew )
        self.actionAbout.triggered.connect( self.slotAbout )
        self.actionShowMultiGitLogFiles.triggered.connect(self.slotShowLogFileDirectory)

        # we enable refresh on focus in
        cast(QApplication, QApplication.instance()).applicationStateChanged.connect(self.slotAppStateChanged)
        self.tabRepos.currentChanged.connect( self.slotCurrentTabChanged )


    def closeEvent(self, event: QCloseEvent) -> None:
        dbg('closeEvent()')
        rect = self.geometry()
        self.config[mgc.CONFIG_MAINWINDOW_GEOMETRY] = (rect.x(), rect.y(), rect.width(), rect.height())

        # save splitter state as a list of bytes, to avoid accessing the QSplitter buffer directly. I suspect
        # that there are sometimes buffer overflow in this buffer, creating bugs in the config file.
        self.config[mgc.CONFIG_SPLITTER_STATE_V2] = [int(v) for v in bytes(self.currentMultigit().splitter.saveState())]

        all_open_tabs = [(self.tabRepos.tabText(tabIdx), cast(MgMultigitWidget, self.tabRepos.widget(tabIdx)).multiRepo.base_dir)
                         for tabIdx in range(self.tabRepos.count())
                         ]
        self.config[mgc.CONFIG_TABS_OPENED] = all_open_tabs
        self.config[mgc.CONFIG_TABS_CURRENT] = self.tabRepos.currentIndex()
        self.config.save()
        return super().closeEvent(event)


    def slotAppStateChanged(self, state: int) -> None:
        if state == Qt.ApplicationActive:
            # window just got activated
            if self.refreshOnFocusBack:
                dbg('slotAppChangedState: %s' % state)
                if self.currentMultigit():
                    selectedItems = self.currentMultigit().repoTree.selectedRepoItems()
                    # only refresh the first item on focus back
                    self.currentMultigit().repoTree.doRefreshItems(selectedItems[:1])


    def checkGitOkAndOpenDefaultRepo(self) -> None:
        '''Check performed on first launch of the UI, to verify that Git is working.'''
        if not ExecGit.checkFound():
            self.slotEditSettingsGitProgram()

        # Open repositories passed as arguments
        if len(self.argv) > 1:
            tabsToOpen = [
                (None, pathlib.Path(argv).resolve().as_posix())
                for argv in self.argv[1:]
            ]
        else:
            tabsToOpen = self.config.lruAsList(mgc.CONFIG_TABS_OPENED)
            if tabsToOpen == []:
                # note: this returns an empty list of the config key does not exist
                #       in this case, we use the last opened directory (previous behavior with no tabs)
                tabsToOpen = [(None, self.config.lruGetFirst(mgc.CONFIG_LAST_OPENED))]

        for tabName, base_dir in reversed(tabsToOpen):
            idx = self.addTab(base_dir, pos=0)
            if tabName:
                self.tabRepos.setTabText(idx, tabName)

        fetchReposOnStartup = self.config[mgc.CONFIG_FETCH_ON_STARTUP]
        if fetchReposOnStartup is None or self.config[mgc.CONFIG_DISPLAY_FETCH_ON_STARTUP_COUNTDOWN] is None:
            self.config[mgc.CONFIG_FETCH_ON_STARTUP] = False
            self.config[mgc.CONFIG_DISPLAY_FETCH_ON_STARTUP_COUNTDOWN] = DISPLAY_FETCH_ON_STARTUP_COUNTDOWN_INIT
            self.config.save()
        else:
            try:
                val = self.config[mgc.CONFIG_DISPLAY_FETCH_ON_STARTUP_COUNTDOWN]
                if val >= 0:
                    val -= 1
                    self.config[mgc.CONFIG_DISPLAY_FETCH_ON_STARTUP_COUNTDOWN] = val
            except TypeError:
                # invalid config value, reset it
                self.config[mgc.CONFIG_DISPLAY_FETCH_ON_STARTUP_COUNTDOWN] = DISPLAY_FETCH_ON_STARTUP_COUNTDOWN_INIT

        if self.config[mgc.CONFIG_DISPLAY_FETCH_ON_STARTUP_COUNTDOWN] == 0:
            answer = QMessageBox.question(self, 'Activate fetch on startup ?', 'Multigit can fetch all your repositories when you launch it. '
                                 'Do you want to activate this behavior ?\n\nNote that you can change it later in the settings dialog.\n')
            fetchReposOnStartup = (answer == QMessageBox.Yes)
            self.config[mgc.CONFIG_FETCH_ON_STARTUP] = fetchReposOnStartup
            self.config.save()

        if fetchReposOnStartup:
            QTimer.singleShot(1000, self.slotFetchAllReposOnAllTabs)



##########################################################################
#
#       Methods related to menu entries NOT acting on selected items
#
##########################################################################

    def slotOpenDir(self) -> None:
        dbg('openDirTriggered()')
        lastdir = self.config.lruGetFirst(mgc.CONFIG_LAST_OPENED) or ''
        newDir = QFileDialog.getExistingDirectory(self, 'Open directory containing git repositories', lastdir)
        if not newDir:
            return

        # pass to current main widget
        self.openDir(newDir)


    def slotCloneFromMgitFile(self) -> None:
        dbg('slotCloneFromMgitFile()')
        return runDialogCloneFromMgitFile(self)


    def slotOpenRecentDirTriggered(self, action: QAction) -> None:
        dbg('openRecentDirTriggered()')
        self.openDir(action.data())


    def updateRecentDirMenu(self, newDir: str) -> None:
        self.config.lruSetRecent(mgc.CONFIG_LAST_OPENED, newDir)
        self.config.save()
        self.fillRecentDirMenu()


    def openDir(self, newDir: str, ignoreNonExisting: bool = False) -> None:
        info('Opening dir: %s' % newDir)
        dirPath = pathlib.Path(newDir)
        if not dirPath.exists():
            if not ignoreNonExisting:
                QMessageBox.warning(self, 'No such directory', 'Can not open directory "%s", it does not exist.' % newDir )
            return
        if not dirPath.is_dir():
            if not ignoreNonExisting:
                QMessageBox.warning(self, 'Not a directory', 'Can not open directory "%s", it is not a directory.' % newDir )
            return

        self.updateRecentDirMenu(newDir)

        # pass to main widget
        self.currentMultigit().setBaseDir(newDir)


    def slotFetchAllReposOnAllTabs(self) -> None:
        '''Called during startup, to fetch really all repositories from all tabs'''
        dbg('slotFetchAllReposOnAllTabs()')

        gitExecWindow = MgExecWindow(self)
        gitExecWindow.setWindowTitle('Fetching all repositories on startup')

        for tabIdx in range(self.tabRepos.count()):
            multiRepo = self.multigitWidgetFromIdx(tabIdx).multiRepo
            cmd = ['fetch', '--prune']
            gitExecWindow.execOneGitCommand(f'Fetching all repositories for {multiRepo.base_dir}', cmd, multiRepo.repo_list)


    def slotExportCsv(self) -> None:
        '''Export to a CSV file'''
        dbg('slotExportCsv()')
        runDialogExportCsv(self, self.currentMultigit().multiRepo)


    def slotExportToMgit(self) -> None:
        dbg('slotExportToMgit()')
        runDialogExportMgit(self, self.currentMultigit().multiRepo.repo_list)


    def slotApplyMgitFile(self) -> None:
        dbg('slotApplyMgitFile()')
        baseDir = self.currentMultigit().multiRepo.base_dir
        allRepos = self.currentMultigit().multiRepo.repo_list[:]
        runDialogApplyMgitFile(self, baseDir, allRepos)


    def slotEditSettings(self) -> None:
        '''Show the Edit Settings dialog'''
        self.editSettings(0)


    def slotEditSettingsGitProgram(self) -> None:
        '''Show the Edit Settings dialog'''
        self.editSettings(1)


    def editSettings(self, startingPage: Union[Literal[0], Literal[1]]) -> None:
        runDialogEditSettings(self, tabPage=startingPage)
        self.mgActions.enableAvailableScm()
        for repoTree in self.allTreeOfMultigitTab():
            repoTree.availableScmUpdated()


    def slotAbout(self) -> None:
        '''Show about dialog'''
        showDialogAbout(self)


    def slotWhatIsNew(self) -> None:
        '''Show what's new dialog'''
        showWhatIsNew(self)


    def selectedItems(self) -> List[MgRepoTreeItem]:
        '''Return the list of selected MgRepoTreeItem'''
        items = cast(List[MgRepoTreeItem], list(self.currentMultigit().repoTree.selectedItems()))
        return items


    # noinspection PyMethodMayBeStatic
    def slotShowLogFileDirectory(self) -> None:
        '''Open an explorer on the log file directory'''
        from src import mg_const
        if mg_const.PATH_LOG_NORMAL is None and mg_const.PATH_LOG_DEBUG is None:
            QMessageBox.warning(None, 'No log file directory', 'MultiGit was unable to use a log file directory!')
            return

        path_log_file = mg_const.PATH_LOG_NORMAL or mg_const.PATH_LOG_DEBUG
        assert path_log_file is not None
        path_log_dir = str(path_log_file.parent)

        if not ExecExplorer.checkFound():
            return
        ExecExplorer.exec_non_blocking([path_log_dir])


    ##########################################################################
    #
    #       Methods to dispatch to or from one or all multigit tabs
    #
    ##########################################################################

    def slotViewTabChanged(self) -> None:
        showTabLastCommit = self.actionViewLastCommit.isChecked()
        showTabModFiles = self.actionViewModifiedFiles.isChecked()
        for tabIdx in range(self.tabRepos.count()):
            self.multigitWidgetFromIdx(tabIdx).slotViewTabChanged(showTabLastCommit, showTabModFiles)

        self.config[mgc.CONFIG_VIEW_TAB_LAST_COMMIT] = showTabLastCommit
        self.config[mgc.CONFIG_VIEW_TAB_MOD_FILES] = showTabModFiles


    def slotViewColSha1Changed(self) -> None:
        '''Called when menu Edit->View->Columns->Sha1 changes'''
        dbg('slotViewColSha1Changed')
        showColSha1 = self.actionViewColSha1.isChecked()
        for repoTree in self.allTreeOfMultigitTab():
            repoTree.slotViewColSha1Changed(showColSha1)
        mgc.get_config_instance()[mgc.CONFIG_VIEW_COL_SHA1] = showColSha1
        mgc.get_config_instance().save()


    def slotViewColUrlChanged(self) -> None:
        '''Called when menu Edit->View->Columns->Sha1 changes'''
        dbg('slotViewColUrlChanged')
        showColUrl = self.actionViewColURL.isChecked()
        for repoTree in self.allTreeOfMultigitTab():
            repoTree.slotViewColUrlChanged(showColUrl)
        mgc.get_config_instance()[mgc.CONFIG_VIEW_COL_URL] = showColUrl
        mgc.get_config_instance().save()


    def slotMenuCopyAboutToShow(self) -> None:
        '''Called when user clicked on the copy submenu. Fills the items of the menu'''
        items = self.currentMultigit().repoTree.selectedRepoItems()
        if len(items) == 0:
            return

        repoInfo = items[0].repoInfo
        self.mgActions.setupDynamicMenuCopy(repoInfo)


    def slotMenuCopyAboutToHide(self) -> None:
        '''Called when Copy menu is going to hide. We disable the connections created
        by the Copy menu to capture the repo missing info'''

        # connections are used to update the menu after is has been shown, when
        # information is missing. We must disconnect everything when the menu
        # goes to hide
        self.mgActions.clearMenuCopyConnections()


    ##########################################################################
    #
    #       Methods to manage tabs
    #
    ##########################################################################

    def slotAddTab(self, pos: int = -1) -> int:
        '''Create a new empty tab, and return the tab index. If pos is not specified or is -1, add the tab
        at the end of the tab list. Else, add it add the expected position.'''
        multigitWidget = MgMultigitWidget()
        if pos == -1:
            pos = self.tabRepos.count()
        tabIdx = self.tabRepos.insertTab(pos, multigitWidget, 'MultiGit')
        self.tabRepos.setCurrentIndex(tabIdx)
        multigitWidget.sig_dir_changed.connect(self.slotTabBaseDirChanged)
        multigitWidget.sig_dir_changed.connect(self.updateRecentDirMenu)
        multigitWidget.sig_request_dir_open.connect( self.slotOpenDir )
        multigitWidget.repoTree.itemSelectionChanged.connect( self.updateStatusBar )
        self.slotViewTabChanged()
        return tabIdx


    def addTab(self, base_dir: Optional[str], pos: int = -1) -> int:
        '''Add a new tab, by default at the end of the tab page. If pos is specified, position of the new tab.
        If base_dir is None, a new empty tab is added

        Return the index of the new tab.
        '''
        tabIdx = self.slotAddTab(pos)
        if base_dir is not None:
            cast(MgMultigitWidget, self.tabRepos.widget(tabIdx)).openDir(base_dir)
        return tabIdx


    def slotDupTab(self) -> None:
        '''Duplicate the current active tab'''
        currentBaseDir = self.currentMultigit().multiRepo.base_dir
        tabIdx = self.slotAddTab()
        self.multigitWidgetFromIdx(tabIdx).openDir( currentBaseDir  )


    def slotRenameTab(self) -> None:
        '''Open a dialog to rename the active tab'''
        curIdx = self.tabRepos.currentIndex()
        oldName = self.tabRepos.tabText(curIdx)
        self.tabRepos.setTabText(curIdx, '')
        tabRenameLineEdit = QLineEdit(oldName)
        self.tabRepos.tabBar().setTabButton(curIdx, QTabBar.LeftSide, tabRenameLineEdit)
        tabRenameLineEdit.selectAll()
        tabRenameLineEdit.setFocus()

        def setNewName() -> None:
            self.tabRepos.setTabText(curIdx, tabRenameLineEdit.text())
            self.tabRepos.tabBar().setTabButton(curIdx, QTabBar.LeftSide, None)     # type: ignore[arg-type]    # the None is not accepted in the stubs yet

        tabRenameLineEdit.returnPressed.connect(setNewName)


    def slotCloseTab(self) -> None:
        '''Close the current tab, unless it is the last one.'''
        if self.tabRepos.count() <= 1:
            QMessageBox.warning(self, 'Can not close tab', 'Can not close last tab!\nOperation canceled.')
            return
        self.tabRepos.removeTab(self.tabRepos.currentIndex())


    def slotCurrentTabChanged(self, tabIdx: int) -> None:
        '''Called when active tab changes'''
        baseDir = self.currentMultigit().multiRepo.base_dir
        self.updateWindowTitleFromBaseDir(baseDir)


    def slotTabBaseDirChanged(self, baseDir: str) -> None:
        '''Called when opening a new base directory in a tab'''
        dirName = pathlib.Path(baseDir).name
        tabidx = self.tabRepos.currentIndex()
        self.tabRepos.setTabToolTip(tabidx, baseDir)
        self.tabRepos.setTabText(tabidx, dirName)
        self.updateWindowTitleFromBaseDir(baseDir)


    def updateWindowTitleFromBaseDir(self, baseDir: str) -> None:
        if len(baseDir):
            self.setWindowTitle('MultiGit OpenSource v{} - {}'.format(VERSION, baseDir))
            self.updateStatusBar()
        else:
            self.setWindowTitle('MultiGit OpenSource v{}'.format(VERSION))
            self.updateStatusBar()


    def updateStatusBar(self) -> None:
        if self.currentMultigit() is None:
            self.statusbar.clearMessage()
            return

        nbRepo = len(self.currentMultigit().multiRepo)
        nbRepoSelected = len(self.currentMultigit().selectedItems())
        msg = f'{nbRepo} repositories, {nbRepoSelected} selected'
        self.statusbar.showMessage(msg, 0)





#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List

import functools, logging

from PySide2.QtCore import QMetaObject, QObject
from PySide2.QtGui import QIcon, QPixmap, QTransform
from PySide2.QtWidgets import QAction, QMenu, QWidget

from src.mg_const import SHORT_SHA1_NB_DIGITS
from src.mg_tools import ExecSublimeMerge, ExecTortoiseGit, ExecSourceTree, ExecGitBash, ExecGitGui, ExecGitK
from src.mg_repo_info import MgRepoInfo

logger = logging.getLogger('mg_actions')
dbg = logger.debug


@functools.lru_cache(maxsize=1)
def getFetchIcon() -> QIcon:
    '''Generate the QIcon for fetch and cache it'''
    # Fetch needs push icon, but rotated
    pixmap_push = QPixmap(":/img/tgit_push.ico")
    transform = QTransform()
    transform.rotate(180)
    pixmap_fetch = pixmap_push.transformed(transform)
    return QIcon(pixmap_fetch)


class MgActions(QObject):
    '''Class to hold all the actions which are duplicated between the main window and each MgRepoTree'''

    actionGitFetchAll: QAction          # added externally to MgActions
    actionGitFetchAllOnAllTabs: QAction          # added externally to MgActions

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.menuCopyConnections = []    # type: List[QMetaObject.Connection]

        ######## Edit actions ##########

        self.actionRefreshSelected = QAction(self)
        self.actionRefreshSelected.setText("Refresh selected repositories")
        self.actionRefreshSelected.setShortcut("Ctrl+F5")
        self.actionRefreshSelected.setIcon(QIcon(":img/tgit_refresh.ico"))
        self.actionRefreshSelected.setToolTip("Refresh selected repositories information from disk")

        self.actionShowInExplorer = QAction(self)
        self.actionShowInExplorer.setText("Show in Explorer")
        self.actionShowInExplorer.setShortcut("Ctrl+E")
        self.actionShowInExplorer.setIcon( QIcon(":img/tgit_explorer.ico"))
        self.actionShowInExplorer.setToolTip('Open Explorer for each repository')

        self.actionCopyFullPath = QAction(self)
        self.actionCopyFullPath.setText('...')
        self.actionCopyFullPath.setToolTip('Copy full path to repository')

        self.actionCopyDirectory = QAction(self)
        self.actionCopyDirectory.setText('...')
        self.actionCopyDirectory.setToolTip('Copy relative directory path to base dir')

        self.actionCopyHead = QAction(self)
        self.actionCopyHead.setText('...')
        self.actionCopyHead.setToolTip('Copy HEAD branch name or tag name')

        self.actionCopyShortSha1 = QAction(self)
        self.actionCopyShortSha1.setText('...')
        self.actionCopyShortSha1.setToolTip('Copy short SHA1 of current commit')

        self.actionCopyLongSha1 = QAction(self)
        self.actionCopyLongSha1.setText('...')
        self.actionCopyLongSha1.setToolTip('Copy full SHA1 of current commit')

        self.actionCopyUrl = QAction(self)
        self.actionCopyUrl.setText('...')
        self.actionCopyUrl.setToolTip('Copy remote URL of repository')


        ######## Git actions ##########

        self.actionGitProperties = QAction(self)
        self.actionGitProperties.setText("Git properties")
        self.actionGitProperties.setShortcut("Alt+Return")
        self.actionGitProperties.setIcon(QIcon(":img/icons8-show-property-96.png"))
        self.actionGitProperties.setToolTip('Open dialog showing all properties of the repository')

        self.actionGitCommit = QAction(self)
        self.actionGitCommit.setText("Commit")
        self.actionGitCommit.setShortcut("Ctrl+Shift+C")
        self.actionGitCommit.setIcon(QIcon(":/img/tgit_commit.ico"))
        self.actionGitCommit.setToolTip('Open dialog to commit on the repositories')

        self.actionGitRevert = QAction(self)
        self.actionGitRevert.setText("Revert modified")
        self.actionGitRevert.setShortcut("Ctrl+R")
        self.actionGitRevert.setIcon(QIcon(":/img/tgit_revert.ico"))
        self.actionGitRevert.setToolTip('Open dialog to revert modified files on repositories')

        self.actionGitPush = QAction(self)
        self.actionGitPush.setText("Push")
        self.actionGitPush.setShortcut("Ctrl+P")
        self.actionGitPush.setIcon(QIcon(":/img/tgit_push.ico"))
        self.actionGitPush.setToolTip('Run <code>git push</code> on repositories')

        self.actionGitPushTag = QAction(self)
        self.actionGitPushTag.setText("Push tag")
        self.actionGitPushTag.setIcon(QIcon(":/img/tgit_push.ico"))
        self.actionGitPushTag.setToolTip('Open dialog to choose which tag to push to remote repository on repositories')

        self.actionGitPull = QAction(self)
        self.actionGitPull.setText("Pull")
        self.actionGitPull.setShortcut("Ctrl+Shift+P")
        self.actionGitPull.setIcon(QIcon(":/img/tgit_pull.ico"))
        self.actionGitPull.setToolTip('Run <code>git pull</code> on repositories')

        self.actionGitFetch = QAction(self)
        self.actionGitFetch.setText("Fetch")
        self.actionGitFetch.setShortcut("Ctrl+F")
        self.actionGitFetch.setIcon(getFetchIcon())
        self.actionGitFetch.setToolTip('Run <code>git fetch</code> on repositories')

        self.actionGitTag = QAction(self)
        self.actionGitTag.setText("Tag")
        self.actionGitTag.setShortcut("Ctrl+T")
        self.actionGitTag.setIcon(QIcon(":/img/tgit_tag.ico"))
        self.actionGitTag.setToolTip('Open dialog for setting tag on repositories')

        self.actionGitCreateBranch = QAction(self)
        self.actionGitCreateBranch.setText("Create branch")
        self.actionGitCreateBranch.setShortcut("Ctrl+B")
        self.actionGitCreateBranch.setToolTip('Open dialog for branch creation')

        self.actionGitSwitchBranch = QAction(self)
        self.actionGitSwitchBranch.setText("Switch to branch")
        self.actionGitSwitchBranch.setShortcut("Ctrl+Shift+S")
        self.actionGitSwitchBranch.setToolTip('Open dialog for switching to a branch')

        self.actionGitCheckoutTag = QAction(self)
        self.actionGitCheckoutTag.setText("Checkout tag")
        # self.actionGitCheckoutTag.setShortcut("Ctrl+Shift+S")
        self.actionGitCheckoutTag.setToolTip('Open dialog for checkouting a tag')

        self.actionGitDeleteBranch = QAction(self)
        self.actionGitDeleteBranch.setText("Delete branch")
        self.actionGitDeleteBranch.setShortcut("Ctrl+Shift+D")
        self.actionGitDeleteBranch.setToolTip('Open dialog for deleting a branch')

        self.actionGitBash = QAction(self)
        self.actionGitBash.setText("Git bash here")
        self.actionGitBash.setIcon(QIcon(":/img/git-bash.ico"))
        self.actionGitBash.setToolTip('Open a git bash window for each repository')

        self.actionGitRunCommand = QAction(self)
        self.actionGitRunCommand.setText("Run git command")
        self.actionGitRunCommand.setShortcut("Ctrl+G")
        self.actionGitRunCommand.setToolTip('Open dialog to run a custom git command')

        ### Git Programs actions
        self.actionConfigureGitProgram = QAction(self)
        self.actionConfigureGitProgram.setEnabled(True)
        self.actionConfigureGitProgram.setText("Open settings to configure a Git Program")

        self.actionSublimeMerge = QAction(self)
        self.actionSublimeMerge.setText("Run SublimeMerge")
        self.actionSublimeMerge.setIcon(QIcon(":/img/sublime_merge.ico"))
        self.actionSublimeMerge.setToolTip('Open a SublimeMerge tab for each repository')

        self.actionSourceTree = QAction(self)
        self.actionSourceTree.setText("Run SourceTree")
        self.actionSourceTree.setIcon(QIcon(":/img/sourcetree.ico"))
        self.actionSourceTree.setToolTip('Open a SourceTree tab for each repository')

        self.actionGitGui = QAction(self)
        self.actionGitGui.setText("Run Git GUI")
        # self.actionGitGui.setIcon(QIcon(":/img/sourcetree.ico"))
        self.actionGitGui.setToolTip('Open a git GUI tab for each repository')

        self.actionGitK = QAction(self)
        self.actionGitK.setText("Run GitK")
        # self.actionGitK.setIcon(QIcon(":/img/sourcetree.ico"))
        self.actionGitK.setToolTip('Open a GitK tab for each repository')

        ######## TortoiseGit actions ##########

        self.actionTGitShowLog = QAction(self)
        self.actionTGitShowLog.setText("Show Log")
        self.actionTGitShowLog.setShortcut("Alt+S")
        self.actionTGitShowLog.setIcon(QIcon(":/img/tgit_log.ico"))
        self.actionTGitShowLog.setToolTip('Open TortoiseGit dialog for Show Log')

        self.actionTGitCommit = QAction(self)
        self.actionTGitCommit.setText("Commit")
        self.actionTGitCommit.setShortcut("Alt+C")
        self.actionTGitCommit.setIcon(QIcon(":/img/tgit_commit.ico"))
        self.actionTGitCommit.setToolTip('Open TortoiseGit dialog for Commit')

        self.actionTGitDiff = QAction(self)
        self.actionTGitDiff.setText("Diff")
        self.actionTGitDiff.setShortcut("Alt+D")
        self.actionTGitDiff.setIcon(QIcon(":/img/tgit_diff.ico"))
        self.actionTGitDiff.setToolTip('Open TortoiseGit dialog for Diff')

        self.actionTGitRevert = QAction(self)
        self.actionTGitRevert.setText("Revert")
        self.actionTGitRevert.setShortcut("Alt+R")
        self.actionTGitRevert.setIcon(QIcon(":/img/tgit_revert.ico"))
        self.actionTGitRevert.setToolTip('Open TortoiseGit dialog for Revert')

        self.actionTGitPush = QAction(self)
        self.actionTGitPush.setText("Push")
        self.actionTGitPush.setShortcut("Alt+P")
        self.actionTGitPush.setIcon(QIcon(":/img/tgit_push.ico"))
        self.actionTGitPush.setToolTip('Open TortoiseGit dialog for Push')

        self.actionTGitPull = QAction(self)
        self.actionTGitPull.setText("Pull")
        self.actionTGitPull.setShortcut("Alt+Shift+P")
        self.actionTGitPull.setIcon(QIcon(":/img/tgit_pull.ico"))
        self.actionTGitPull.setToolTip('Open TortoiseGit dialog for Pull')

        self.actionTGitFetch = QAction(self)
        self.actionTGitFetch.setText("Fetch")
        self.actionTGitFetch.setShortcut("Alt+F")
        self.actionTGitFetch.setIcon(getFetchIcon())
        self.actionTGitFetch.setToolTip('Open TortoiseGit dialog for Fetch')

        self.actionTGitTag = QAction(self)
        self.actionTGitTag.setText("Tag")
        self.actionTGitTag.setShortcut("Alt+T")
        self.actionTGitTag.setIcon(QIcon(":/img/tgit_tag.ico"))
        self.actionTGitTag.setToolTip('Open TortoiseGit dialog for Tag')

        self.actionTGitSwitch = QAction(self)
        self.actionTGitSwitch.setText("Switch")
        self.actionTGitSwitch.setShortcut("Alt+W")
        self.actionTGitSwitch.setIcon(QIcon(":/img/tgit_switch.ico"))
        self.actionTGitSwitch.setToolTip('Open TortoiseGit dialog for Switch')

        self.actionTGitBranch = QAction(self)
        self.actionTGitBranch.setText("Branch")
        self.actionTGitBranch.setShortcut("Alt+B")
        self.actionTGitBranch.setIcon(QIcon(":/img/tgit_branch.ico"))
        self.actionTGitBranch.setToolTip('Open TortoiseGit dialog for Branch')


    def setupMenuView(self, menuView: QMenu) -> None:
        '''Install more actions in the menu View'''
        menuView.addAction(self.actionShowInExplorer)
        menuView.addAction(self.actionRefreshSelected)
        self.menuCopy = QMenu('Copy', menuView)
        menuView.addMenu(self.menuCopy)

        '''Setup the internal menu copy used in menu edit'''
        self.menuCopy.setToolTip('Copy repository information to clipboard')
        self.menuCopy.setToolTipsVisible(True)
        self.menuCopy.addAction(self.actionCopyFullPath)
        self.menuCopy.addAction(self.actionCopyDirectory)
        self.menuCopy.addAction(self.actionCopyHead)
        self.menuCopy.addAction(self.actionCopyShortSha1)
        self.menuCopy.addAction(self.actionCopyLongSha1)
        self.menuCopy.addAction(self.actionCopyUrl)



    def setupMenuGit(self, menuGit: QMenu) -> None:
        '''Install all actions in the git menu'''
        menuGit.addAction( self.actionGitProperties)
        menuGit.addAction( self.actionGitCommit )
        menuGit.addAction( self.actionGitRevert )
        menuGit.addAction( self.actionGitPush )
        menuGit.addAction( self.actionGitPushTag )
        menuGit.addAction( self.actionGitPull )
        menuGit.addAction( self.actionGitFetch )
        menuGit.addAction( self.actionGitTag )
        menuGit.addAction( self.actionGitCreateBranch )
        menuGit.addAction( self.actionGitSwitchBranch)
        menuGit.addAction( self.actionGitCheckoutTag)
        menuGit.addAction( self.actionGitDeleteBranch)
        menuGit.addSeparator()
        menuGit.addAction( self.actionGitBash )
        menuGit.addAction( self.actionGitRunCommand )


    def setupMenuGitPrograms(self, menuGitPrograms: QMenu, includeConfigureGitPrograms: bool) -> None:
        '''Install more actions in the GitPrograms menu'''
        menuGitPrograms.addAction(self.actionSourceTree)
        menuGitPrograms.addAction(self.actionSublimeMerge)
        menuGitPrograms.addAction(self.actionGitGui)
        menuGitPrograms.addAction(self.actionGitK)
        self.menuTGit = menuGitPrograms.addMenu("Tortoise Git")
        self.menuTGit.setIcon(QIcon(":/img/tgit_tortoise.ico"))
        self.setupMenuTortoiseGit(self.menuTGit)
        if includeConfigureGitPrograms:
            menuGitPrograms.addAction(self.actionConfigureGitProgram)
        self.enableAvailableScm()


    def setupMenuTortoiseGit(self, menuTGit: QMenu) -> None:
        '''Install all actions in the TortoiseGit menu'''
        menuTGit.addAction( self.actionTGitShowLog )
        menuTGit.addAction( self.actionTGitCommit )
        menuTGit.addAction( self.actionTGitDiff )
        menuTGit.addAction( self.actionTGitRevert )
        menuTGit.addAction( self.actionTGitPush )
        menuTGit.addAction( self.actionTGitPull )
        menuTGit.addAction( self.actionTGitFetch )
        menuTGit.addAction( self.actionTGitTag )
        menuTGit.addAction( self.actionTGitSwitch )
        menuTGit.addAction( self.actionTGitBranch )


    def enableAvailableScm(self) -> None:
        '''Adjust Git Programs menu and RMB menu'''
        showTortoiseGit = ExecTortoiseGit.shouldShow()
        showSourceTree = ExecSourceTree.shouldShow()
        showSublimeMerge = ExecSublimeMerge.shouldShow()
        showGitBash = ExecGitBash.shouldShow()
        showGitGui = ExecGitGui.shouldShow()
        showGitK = ExecGitK.shouldShow()

        # fix menu actions
        self.actionSourceTree.setVisible(showSourceTree)
        self.actionSublimeMerge.setVisible(showSublimeMerge)
        self.actionGitBash.setVisible(showGitBash)
        self.actionGitGui.setVisible(showGitGui)
        self.actionGitK.setVisible(showGitK)
        self.menuTGit.menuAction().setVisible(showTortoiseGit)

        if showTortoiseGit or showSourceTree or showSublimeMerge or showGitGui or showGitK:
            self.actionConfigureGitProgram.setVisible(False)
        else:
            self.actionConfigureGitProgram.setVisible(True)


    def setupDynamicMenuCopy(self, repoInfo: MgRepoInfo) -> None:
        self.actionCopyFullPath.setText(repoInfo.fullpath)
        self.actionCopyDirectory.setText(repoInfo.name)

        if repoInfo.head:
            self.actionCopyHead.setText(repoInfo.head.split(' ')[1])
            self.actionCopyHead.setEnabled(True)
        else:
            self.actionCopyHead.setText('...head...')
            self.actionCopyHead.setEnabled(False)
            # if head is missing, this is because the refresh has not completed yet
            # we just have to wait
            def local_set_head(_repoName: str) -> None:
                self.actionCopyHead.setText(repoInfo.head.split(' ')[1])
                self.actionCopyHead.setEnabled(True)
            self.menuCopyConnections.append( repoInfo.repo_info_available.connect(local_set_head) ) # type: ignore

        if repoInfo.commit_sha1:
            self.actionCopyShortSha1.setText(repoInfo.commit_sha1[:SHORT_SHA1_NB_DIGITS])
            self.actionCopyShortSha1.setEnabled(True)
            self.actionCopyLongSha1.setText(repoInfo.commit_sha1)
            self.actionCopyLongSha1.setEnabled(True)
        else:
            self.actionCopyShortSha1.setText('...short sha1...')
            self.actionCopyShortSha1.setEnabled(False)
            self.actionCopyLongSha1.setText('...sha1...')
            self.actionCopyLongSha1.setEnabled(False)
            # if sha1 is missing, we probably need to request it
            def local_set_sha1(_sha1: str) -> None:
                self.actionCopyShortSha1.setText((repoInfo.commit_sha1 or '')[:SHORT_SHA1_NB_DIGITS])
                self.actionCopyShortSha1.setEnabled(True)
                self.actionCopyLongSha1.setText(repoInfo.commit_sha1 or '')
                self.actionCopyLongSha1.setEnabled(True)
            repoInfo.ensure_sha1(local_set_sha1)

        if repoInfo.url is not None:
            self.actionCopyUrl.setText(repoInfo.url)
            self.actionCopyUrl.setEnabled(True)
        else:
            self.actionCopyUrl.setText('...url...')
            self.actionCopyUrl.setEnabled(False)
            def local_set_url(url: str) -> None:
                self.actionCopyUrl.setText(url)
                self.actionCopyUrl.setEnabled(True)
            repoInfo.ensure_url(local_set_url)


    def clearMenuCopyConnections(self) -> None:
        '''Called when Copy menu is going to hide. We disable the connections created
        by the Copy menu to capture the repo missing info'''
        for c in self.menuCopyConnections:
            if bool(c):
                # QObject.disconnect(c)
                # this is not working in pyside2
                # re-enable it in pyside6
                pass
        self.menuCopyConnections = []







# -*- coding: utf-8 -*-

#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


# Form implementation generated from reading ui file 'ui_git_delete_branch.ui'
#
# Created by: PyQt5 UI code generator 5.13.2
# Patched by patch_ui.py
#
# WARNING! All changes made in this file will be lost!


from PySide2 import QtCore, QtGui, QtWidgets


class Ui_GitDeleteBranch(object):
    def setupUi(self, GitDeleteBranch: QtWidgets.QDialog) -> None:
        GitDeleteBranch.setObjectName("GitDeleteBranch")
        GitDeleteBranch.resize(883, 423)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(GitDeleteBranch)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.groupBox = QtWidgets.QGroupBox(GitDeleteBranch)
        self.groupBox.setObjectName("groupBox")
        self.gridLayout = QtWidgets.QGridLayout(self.groupBox)
        self.gridLayout.setObjectName("gridLayout")
        spacerItem = QtWidgets.QSpacerItem(554, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 3, 1, 1)
        self.comboBoxBranchName = QtWidgets.QComboBox(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(3)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBoxBranchName.sizePolicy().hasHeightForWidth())
        self.comboBoxBranchName.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setFamily("Consolas")
        self.comboBoxBranchName.setFont(font)
        self.comboBoxBranchName.setEditable(True)
        self.comboBoxBranchName.setObjectName("comboBoxBranchName")
        self.gridLayout.addWidget(self.comboBoxBranchName, 0, 1, 1, 2)
        spacerItem1 = QtWidgets.QSpacerItem(554, 20, QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem1, 1, 3, 1, 1)
        self.label = QtWidgets.QLabel(self.groupBox)
        self.label.setLayoutDirection(QtCore.Qt.RightToLeft)
        self.label.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)
        self.checkBoxDeleteLR = QtWidgets.QCheckBox(self.groupBox)
        self.checkBoxDeleteLR.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.checkBoxDeleteLR.setObjectName("checkBoxDeleteLR")
        self.gridLayout.addWidget(self.checkBoxDeleteLR, 1, 1, 1, 2)
        self.verticalLayout_2.addWidget(self.groupBox)
        self.groupBox_2 = QtWidgets.QGroupBox(GitDeleteBranch)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.labelRepoSelected = QtWidgets.QLabel(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelRepoSelected.sizePolicy().hasHeightForWidth())
        self.labelRepoSelected.setSizePolicy(sizePolicy)
        self.labelRepoSelected.setAlignment(QtCore.Qt.AlignCenter)
        self.labelRepoSelected.setObjectName("labelRepoSelected")
        self.horizontalLayout.addWidget(self.labelRepoSelected)
        self.pushButtonAdjustRepoList = QtWidgets.QPushButton(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButtonAdjustRepoList.sizePolicy().hasHeightForWidth())
        self.pushButtonAdjustRepoList.setSizePolicy(sizePolicy)
        self.pushButtonAdjustRepoList.setObjectName("pushButtonAdjustRepoList")
        self.horizontalLayout.addWidget(self.pushButtonAdjustRepoList)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.treeWidgetRepoList = QtWidgets.QTreeWidget(self.groupBox_2)
        self.treeWidgetRepoList.setAlternatingRowColors(True)
        self.treeWidgetRepoList.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.treeWidgetRepoList.setRootIsDecorated(False)
        self.treeWidgetRepoList.setUniformRowHeights(True)
        self.treeWidgetRepoList.setItemsExpandable(False)
        self.treeWidgetRepoList.setObjectName("treeWidgetRepoList")
        self.verticalLayout.addWidget(self.treeWidgetRepoList)
        self.verticalLayout_2.addWidget(self.groupBox_2)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.buttonBox = QtWidgets.QDialogButtonBox(GitDeleteBranch)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.horizontalLayout_4.addWidget(self.buttonBox)
        self.verticalLayout_2.addLayout(self.horizontalLayout_4)

        self.retranslateUi(GitDeleteBranch)
        self.buttonBox.accepted.connect(GitDeleteBranch.accept)
        self.buttonBox.rejected.connect(GitDeleteBranch.reject)
        QtCore.QMetaObject.connectSlotsByName(GitDeleteBranch)

    def retranslateUi(self, GitDeleteBranch: QtWidgets.QDialog) -> None:
        _translate = QtCore.QCoreApplication.translate
        GitDeleteBranch.setWindowTitle(_translate("GitDeleteBranch", "Delete branch"))
        self.groupBox.setTitle(_translate("GitDeleteBranch", "Delete Branch"))
        self.label.setText(_translate("GitDeleteBranch", "Enter branch name :"))
        self.checkBoxDeleteLR.setText(_translate("GitDeleteBranch", "Delete locally and remotely"))
        self.groupBox_2.setTitle(_translate("GitDeleteBranch", "Targeted repositories"))
        self.labelRepoSelected.setText(_translate("GitDeleteBranch", "3 repositories selected"))
        self.pushButtonAdjustRepoList.setText(_translate("GitDeleteBranch", "Adjust Repository list"))
        self.treeWidgetRepoList.setSortingEnabled(True)
        self.treeWidgetRepoList.headerItem().setText(0, _translate("GitDeleteBranch", "hidden"))
        self.treeWidgetRepoList.headerItem().setText(1, _translate("GitDeleteBranch", "Git Repo Path"))
        self.treeWidgetRepoList.headerItem().setText(2, _translate("GitDeleteBranch", "Head"))
        self.treeWidgetRepoList.headerItem().setText(3, _translate("GitDeleteBranch", "Status"))
        self.treeWidgetRepoList.headerItem().setText(4, _translate("GitDeleteBranch", "Last Remote Syncho"))



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'ui_git_revert.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from src.mg_repo_tree import MgRepoTree


class Ui_GitRevert(object):
    def setupUi(self, GitRevert: QDialog) -> None:
        if not GitRevert.objectName():
            GitRevert.setObjectName(u"GitRevert")
        GitRevert.resize(768, 561)
        self.verticalLayout_3 = QVBoxLayout(GitRevert)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.groupBox = QGroupBox(GitRevert)
        self.groupBox.setObjectName(u"groupBox")
        self.verticalLayout_2 = QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.label = QLabel(self.groupBox)
        self.label.setObjectName(u"label")
        self.label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)

        self.verticalLayout_2.addWidget(self.label)

        self.textEditRevertContent = QTextEdit(self.groupBox)
        self.textEditRevertContent.setObjectName(u"textEditRevertContent")

        self.verticalLayout_2.addWidget(self.textEditRevertContent)


        self.verticalLayout_3.addWidget(self.groupBox)

        self.widget = QWidget(GitRevert)
        self.widget.setObjectName(u"widget")
        self.widget.setMinimumSize(QSize(0, 30))

        self.verticalLayout_3.addWidget(self.widget)

        self.groupBox_2 = QGroupBox(GitRevert)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.verticalLayout = QVBoxLayout(self.groupBox_2)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.labelRepoSelected = QLabel(self.groupBox_2)
        self.labelRepoSelected.setObjectName(u"labelRepoSelected")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelRepoSelected.sizePolicy().hasHeightForWidth())
        self.labelRepoSelected.setSizePolicy(sizePolicy)
        self.labelRepoSelected.setAlignment(Qt.AlignCenter)

        self.horizontalLayout.addWidget(self.labelRepoSelected)

        self.pushButtonAdjustRepoList = QPushButton(self.groupBox_2)
        self.pushButtonAdjustRepoList.setObjectName(u"pushButtonAdjustRepoList")
        sizePolicy1 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(1)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.pushButtonAdjustRepoList.sizePolicy().hasHeightForWidth())
        self.pushButtonAdjustRepoList.setSizePolicy(sizePolicy1)

        self.horizontalLayout.addWidget(self.pushButtonAdjustRepoList)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.treeWidgetRepoList = MgRepoTree(self.groupBox_2)
        self.treeWidgetRepoList.setObjectName(u"treeWidgetRepoList")
        self.treeWidgetRepoList.setAlternatingRowColors(True)
        self.treeWidgetRepoList.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeWidgetRepoList.setRootIsDecorated(False)
        self.treeWidgetRepoList.setUniformRowHeights(True)
        self.treeWidgetRepoList.setItemsExpandable(False)
        self.treeWidgetRepoList.setSortingEnabled(True)

        self.verticalLayout.addWidget(self.treeWidgetRepoList)


        self.verticalLayout_3.addWidget(self.groupBox_2)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.buttonBox = QDialogButtonBox(GitRevert)
        self.buttonBox.setObjectName(u"buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Yes)

        self.horizontalLayout_4.addWidget(self.buttonBox)


        self.verticalLayout_3.addLayout(self.horizontalLayout_4)

        QWidget.setTabOrder(self.textEditRevertContent, self.pushButtonAdjustRepoList)
        QWidget.setTabOrder(self.pushButtonAdjustRepoList, self.treeWidgetRepoList)

        self.retranslateUi(GitRevert)
        self.buttonBox.accepted.connect(GitRevert.accept)
        self.buttonBox.rejected.connect(GitRevert.reject)

        QMetaObject.connectSlotsByName(GitRevert)
    # setupUi

    def retranslateUi(self, GitRevert: QDialog) -> None:
        GitRevert.setWindowTitle(QCoreApplication.translate("GitRevert", u"Git Create branch", None))
        self.groupBox.setTitle(QCoreApplication.translate("GitRevert", u"Git Revert", None))
        self.label.setText(QCoreApplication.translate("GitRevert", u"Revert the following changes :", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("GitRevert", u"Targeted repositories", None))
        self.labelRepoSelected.setText(QCoreApplication.translate("GitRevert", u"3 repositories selected", None))
        self.pushButtonAdjustRepoList.setText(QCoreApplication.translate("GitRevert", u"Adjust Repository list", None))
        ___qtreewidgetitem = self.treeWidgetRepoList.headerItem()
        ___qtreewidgetitem.setText(4, QCoreApplication.translate("GitRevert", u"Last Remote Syncho", None));
        ___qtreewidgetitem.setText(3, QCoreApplication.translate("GitRevert", u"Status", None));
        ___qtreewidgetitem.setText(2, QCoreApplication.translate("GitRevert", u"Head", None));
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("GitRevert", u"Git Repo Path", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("GitRevert", u"hidden", None));
    # retranslateUi




#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


import logging
from typing import TYPE_CHECKING, cast, Type, Union, Optional, List, Protocol

from PySide2.QtWidgets import QMessageBox, QDialog, QTreeWidget, QAbstractItemView, QPushButton, QLabel, QWidget
from PySide2.QtCore import Qt, Signal

if TYPE_CHECKING:
    from src.mg_window import MgMainWindow
    from src.mg_repo_tree import MgRepoTree
from src.mg_repo_info import MgRepoInfo
from src.mg_repo_tree_item import MgRepoTreeItem
from src.mg_const import COL_NB, COL_TITLES, COL_REPO_NAME, COL_UPDATE

logger = logging.getLogger('mg_dialog_utils')
dbg = logger.debug

reBranchTagValues = "[^ ~^:]*"


if TYPE_CHECKING:
    class UiOfDialogWithRepos(Protocol):
        '''Typing helper class to declare that we need an object with the follwing attributes:
        - has a setupUI() method
        - has a treeWidgetRepoList attribute of type MgRepoTree
        '''
        # typing helper class
        def setupUi(self, dialog: QDialog) -> None:
            ...

        treeWidgetRepoList: 'MgRepoTree'
        pushButtonAdjustRepoList: QPushButton
        labelRepoSelected: QLabel


def prepareTreeWidgetRepoList(treeWidget: QTreeWidget) -> None:
    '''Prepare the tree widget headers and other attributes'''
    if treeWidget.columnCount() < COL_NB:
        treeWidget.setColumnCount(COL_NB)
        treeWidget.setHeaderLabels(COL_TITLES)
        treeWidget.setSelectionMode(QAbstractItemView.ExtendedSelection)
        treeWidget.setAllColumnsShowFocus(True)
        treeWidget.header().setSortIndicatorShown(True)
    f = treeWidget.header().font()
    f.setBold(True)
    treeWidget.header().setFont(f)
    treeWidget.sortByColumn(COL_REPO_NAME, Qt.AscendingOrder)
    treeWidget.headerItem().setText(COL_UPDATE, "")
    treeWidget.setColumnHidden(COL_UPDATE, True)
    treeWidget.clear()


class MgDialogWithRepoList(QDialog):
    '''Inherit from this dialog to manage a dialog with a list of Repositories'''

    sigRepoListAdjusted = Signal()

    def __init__(self, parent: QWidget, Ui_Class: Type['UiOfDialogWithRepos'],
                 selectedRepos: Optional[List[MgRepoInfo]] = None, allRepos: Optional[List[MgRepoInfo]] = None) -> None:
        '''This __init__ method has two ways of being used:
        * new style:
            __init__(ui_class, parentWidget, [repos, ...], [repos, ...])

        Eventually, the legacy style should disappear

        '''
        super().__init__(parent)

        self.ui = Ui_Class()
        self.ui.setupUi(self)

        self.allRepos = allRepos or []
        selectedRepos = selectedRepos or []

        assert parent is not None
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        # prepare repository list
        prepareTreeWidgetRepoList(self.ui.treeWidgetRepoList)

        # fill the list of repositories
        for repo in selectedRepos:
            repoTreeItem = MgRepoTreeItem(repo, self.ui.treeWidgetRepoList)
            repoTreeItem.fillRepoItem()

        MgRepoTreeItem.autoAdjustColumnSize(self.ui.treeWidgetRepoList)

        self.ui.pushButtonAdjustRepoList.clicked.connect(self.slotAdjutRepoList)
        self.updateRepoSelectedLabel()
        self.allReposStatus = "all"


    def setAllReposStatus(self, status:str) -> None:
        self.allReposStatus = status


    def updateRepoSelectedLabel(self) -> None:
        self.ui.labelRepoSelected.setText( '%d repositories targeted' % self.ui.treeWidgetRepoList.topLevelItemCount())


    def getTargetedRepoList(self) -> List[MgRepoInfo]:
        '''Return a list of MgRepoInfo items'''
        targetedRepos = [cast(MgRepoTreeItem, self.ui.treeWidgetRepoList.topLevelItem(idx)).repoInfo
                         for idx in range(self.ui.treeWidgetRepoList.topLevelItemCount())]
        return targetedRepos


    def slotAdjutRepoList(self) -> None:
        # to avoid a circular dependency, perform a local import
        from src.mg_dialog_select_repo import MgDialogSelectRepos
        dialog = MgDialogSelectRepos(self, self.getTargetedRepoList(), self.allRepos)

        result = dialog.exec()
        if not result:
            # dialog canceled
            return

        # adjust list of repos to operate on
        itemToRemove = []
        selectedRepoInfo = dialog.selectedRepoInfo()
        for idx in range(self.ui.treeWidgetRepoList.topLevelItemCount()):
            repoItem = cast(MgRepoTreeItem, self.ui.treeWidgetRepoList.topLevelItem(idx))
            if repoItem.repoInfo not in selectedRepoInfo:
                itemToRemove.append(idx)

        for idx in reversed(itemToRemove):
            self.ui.treeWidgetRepoList.takeTopLevelItem(idx)

        currentlySelectedRepoInfo = [
            cast(MgRepoTreeItem, self.ui.treeWidgetRepoList.topLevelItem(idx)).repoInfo
            for idx in range(self.ui.treeWidgetRepoList.topLevelItemCount())
        ]

        repoToAdd = set(selectedRepoInfo) - set(currentlySelectedRepoInfo)
        for repoInfo in repoToAdd:
            repoTreeItem = MgRepoTreeItem(repoInfo, self.ui.treeWidgetRepoList)
            repoTreeItem.fillRepoItem()

        self.updateRepoSelectedLabel()
        # noinspection PyUnresolvedReferences
        self.sigRepoListAdjusted.emit()


    def accept(self) -> None:
        if self.ui.treeWidgetRepoList.topLevelItemCount() == 0:
            # no item in the list !
            QMessageBox.warning(self, "No repository",
                                "No repository selected.\nYou must select at least one repository!")
            return

        super().accept()


#    Copyright (c) 2019-2023 IDEMIA
#    Author: IDEMIA (Philippe Fremy, Florent Oulieres)
# 
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
# 
#         http://www.apache.org/licenses/LICENSE-2.0
# 
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
#


from typing import List, Any, TYPE_CHECKING
import logging

from PySide2.QtWidgets import QDialog, QWidget, QFileDialog, QMessageBox
from PySide2.QtCore import Qt, QTimer

if TYPE_CHECKING:
    from src.mg_window import MgMainWindow
from src.gui.ui_export_mgit import Ui_ExportMgit
from src.mg_repo_info import MgRepoInfo
from src.mg_ensure_info_available import MgEnsureInfoAvailable, RepoInfoFlags
from src.mg_json_mgit_parser import exportToMgit
from src.mg_utils import add_suffix_if_missing
import src.mg_config as mgc
from src.mg_const import MSG_EMPTY_REPO, MSG_LOCAL_BRANCH, MSG_REMOTE_BRANCH_GONE

logger = logging.getLogger('mg_dialog_export_mgit')
dbg = logger.debug
warn = logger.warning


class MgDialogExportMgit(QDialog):

    def __init__(self, window: QWidget, allRepos: List[MgRepoInfo]) -> None:
        super().__init__(window)
        # noinspection PyTypeChecker
        self.setWindowFlags( self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self.ui = Ui_ExportMgit()
        self.ui.setupUi(self)

        self.mgitFname = ''
        self.exportAsSnapshot = False
        self.allRepos = allRepos

        # set last values
        self.ui.historyButtonMgitFile.fillHistory(mgc.get_config_instance().lruAsList(mgc.CONFIG_LAST_MGIT_FILE))
        self.ui.lineEditMgitFile.setText('')

        # connect buttons to their slot
        self.ui.pushButtonChooseMgitFile.clicked.connect(self.slotChooseMgitFile)
        self.ui.historyButtonMgitFile.historyItemTriggered.connect(self.slotMgitFileHistoryTriggered)
        self.ui.pushButtonExportProject.clicked.connect(self.slotExportAsProject)
        self.ui.pushButtonExportSnapshot.clicked.connect(self.slotExportAsSnapshot)
        self.ui.pushButtonCancel.clicked.connect(self.reject)

        QTimer.singleShot(0, self.slotEnsureRepoFilled)


    def slotEnsureRepoFilled(self) -> None:
        '''While showing the dialog, start filling all repos'''
        self.ensureInfoAvailable = MgEnsureInfoAvailable(self, self.allRepos)
        # request all information in advance while the user is manipulating the dialog
        self.ensureInfoAvailable.ensureInfoAvailable(RepoInfoFlags.HEAD | RepoInfoFlags.URL, blocking=False)
        # call it in a second step because maybe we don't need it
        self.ensureInfoAvailable.ensureInfoAvailable(RepoInfoFlags.SHA1, blocking=False)


    def slotMgitFileHistoryTriggered(self, title: str, _message: str) -> None:
        '''History of Mgit file triggered, open the targeted mgit file'''
        self.ui.lineEditMgitFile.setText(title)


    def slotChooseMgitFile(self) -> None:
        """
        function called when pushing button to select configuration files
        """
        dbg('slotChooseMgitFile')

        mgitFile = self.ui.lineEditMgitFile.text()

        # ask user to select configuration file
        lastdir = mgc.get_config_instance().lruGetFirst(mgc.CONFIG_LAST_OPENED) or ''

        mgitFile, _ = QFileDialog.getSaveFileName(self, "Select destination multigit file",
                    mgitFile or lastdir, "Multigit files (*.mgit)")

        if not mgitFile:
            # dialog canceled, do nothing
            return

        self.ui.lineEditMgitFile.setText(mgitFile)


    def slotExportAsProject(self) -> None:
        '''Called when exporting as project'''
        self.exportAsSnapshot = False
        self.accept()


    def slotExportAsSnapshot(self) -> None:
        '''Called when exporting as snapshot'''
        self.exportAsSnapshot = True
        self.accept()


    def accept(self) -> None:
        '''Called when the dialog is validated'''
        if self.ui.lineEditMgitFile.text() == '':
            QMessageBox.warning(self, 'No multigit file', 'You must choose a destination multigit file.')
            return

        # We had a bug report once that a repo may have been deleted in the meantime. This is
        # confusing for the user because the deletion of repo updated the main window with the list of repositories
        # and the list of repo is updated by removing the repo. So, the user legitimately believes that the deleted
        # directory is not included in the export. However, the list we use here has not yet been updated.
        # So, add a warning for this case.

        # filter directly without warning if the repo no longer exists
        for repo in self.allRepos:
            if repo.is_deleted:
                dbg(f'Removing deleted repo from list: {repo.name}')
        self.allRepos = [repo for repo in self.allRepos if not repo.is_deleted ]


        # we want to have the URL available
        ensureInfoAvailable = MgEnsureInfoAvailable(self, self.allRepos, showProgressDialog=False)
        ensureInfoAvailable.ensureInfoAvailable(RepoInfoFlags.URL, blocking=True)

        repoWithEmtpyUrl = [repo.name for repo in self.allRepos if repo.url == '']
        if len(repoWithEmtpyUrl):
            msg = '''The following repositories have no remote URL defined and can not be used for a future 
            clone operation:<ul>\n<li>'''
            msg += '\n<li> '.join(repoWithEmtpyUrl)
            msg += '''</ul>What do you want to do ?<p>'''
            msgBox = QMessageBox(QMessageBox.Icon.Warning, 'Repositories with empty remote URL', msg)
            msgBox.setTextFormat(Qt.RichText)
            buttonCancel = msgBox.addButton(QMessageBox.Cancel)
            buttonSkip = msgBox.addButton('Skip repositories', QMessageBox.AcceptRole)
            buttonContinue = msgBox.addButton('Continue', QMessageBox.AcceptRole)
            msgBox.exec()
            buttonSelected = msgBox.clickedButton()
            if buttonSelected == buttonCancel:
                return
            elif buttonSelected == buttonSkip:
                # adjust the list of repositories before accepting the dialog
                self.allRepos = [repo for repo in self.allRepos if repo.url != '']

        repoWithLocalOnlyBranch = [repo for repo in self.allRepos
                                   if repo.branch != '' and repo.remote_synchro in (MSG_REMOTE_BRANCH_GONE, MSG_LOCAL_BRANCH)]
        if len(repoWithLocalOnlyBranch):
            msg = '''Some repositories are on a branch which does not exist remotely. The clone operation 
            from this multigit file will fail. The repositories are:<ul><li>'''
            msg += '\n<li> '.join(f'Repository {repo.name} - branch {repo.branch}' for repo in repoWithLocalOnlyBranch)
            msg += '''</ul>What do you want to do ?<p>'''
            msgBox = QMessageBox(QMessageBox.Icon.Warning, 'Repositories with local only branch', msg)
            msgBox.setTextFormat(Qt.RichText)
            msgBox.setText(msg)
            buttonCancel = msgBox.addButton(QMessageBox.Cancel)
            buttonContinue = msgBox.addButton('Continue', QMessageBox.AcceptRole)
            msgBox.exec()
            buttonSelected = msgBox.clickedButton()
            if buttonSelected == buttonCancel:
                return

        self.mgitFname = add_suffix_if_missing(self.ui.lineEditMgitFile.text(), '.mgit')

        mgc.get_config_instance().lruSetRecent(mgc.CONFIG_LAST_MGIT_FILE, self.mgitFname)
        mgc.get_config_instance().save()
        # everyting is fine, accept the dialog
        super().accept()


def runDialogExportMgit(window: QWidget, allRepos: List[MgRepoInfo]) -> None:
    """
    Dialog to export current repository state to a multigit file.
    """
    dbg('runDialogExportMgit')
    if len(allRepos) == 0:
        QMessageBox.warning(window, 'No repository defined',
                            'Warning: can not export when base directory is not defined.')
        return

    dialogExportMgit = MgDialogExportMgit(window, allRepos)

    # display dialog
    dlg_result = dialogExportMgit.exec()
    if not dlg_result:
        # Cancel was clicked, do nothing
        return

    # now, ensure all information is really available with a progress bar before calling exportToMgit()
    ensureInfoAvailable = MgEnsureInfoAvailable(dialogExportMgit, dialogExportMgit.allRepos, showProgressDialog=True)
    infoRequested = RepoInfoFlags.HEAD | RepoInfoFlags.URL
    if dialogExportMgit.exportAsSnapshot:
        infoRequested |= RepoInfoFlags.SHA1

    ensureInfoAvailable.ensureInfoAvailable(infoRequested, blocking=True)


    exportToMgit(dialogExportMgit.mgitFname,
                 dialogExportMgit.ui.lineEditDescMgit.text(),
                 dialogExportMgit.allRepos, snapshotMode=dialogExportMgit.exportAsSnapshot)


    QMessageBox.information(window, 'Information exported',
                            'Export done to %s' % dialogExportMgit.mgitFname,
                            QMessageBox.Ok)
